from __future__ import annotations

import base64
import io
import logging
import os
import platform
from typing import TYPE_CHECKING

from ..browser.views import PLACEHOLDER_4PX_SCREENSHOT
from ..config import CONFIG
from .views import AgentHistoryList

if TYPE_CHECKING:
    from PIL import Image, ImageFont

logger = logging.getLogger(__name__)


def decode_unicode_escapes_to_utf8(text: str) -> str:
    """Handle decoding any unicode escape sequences embedded in a string.

    Needed to render non-ASCII languages like Chinese or Arabic in the GIF overlay text.
    """

    if r"\u" not in text:
        # doesn't have any escape sequences that need to be decoded
        return text

    try:
        # Try to decode Unicode escape sequences
        return text.encode("latin1").decode("unicode_escape")
    except (UnicodeEncodeError, UnicodeDecodeError):
        # logger.debug(f"Failed to decode unicode escape sequences while generating gif text: {text}")
        return text


def create_history_gif(
    task: str,
    history: AgentHistoryList,
    #
    output_path: str = "agent_history.gif",
    duration: int = 3000,
    show_goals: bool = True,
    show_task: bool = True,
    show_logo: bool = False,
    font_size: int = 40,
    title_font_size: int = 56,
    goal_font_size: int = 44,
    margin: int = 40,
    line_spacing: float = 1.5,
) -> None:
    """Create a GIF from the agent's history with overlaid task and goal text."""
    if not history.history:
        logger.warning("No history to create GIF from")
        return

    from PIL import Image, ImageFont

    images = []

    # if history is empty, we can't create a gif
    if not history.history:
        logger.warning("No history to create GIF from")
        return

    # Get all screenshots from history (including None placeholders)
    screenshots = history.screenshots(return_none_if_not_screenshot=True)

    if not screenshots:
        logger.warning("No screenshots found in history")
        return

    # Find the first non-placeholder screenshot
    # A screenshot is considered a placeholder if:
    # 1. It's the exact 4px placeholder for about:blank pages, OR
    # 2. It comes from a new tab page (chrome://newtab/, about:blank, etc.)
    first_real_screenshot = None
    for screenshot in screenshots:
        if screenshot and screenshot != PLACEHOLDER_4PX_SCREENSHOT:
            first_real_screenshot = screenshot
            break

    if not first_real_screenshot:
        logger.warning("No valid screenshots found (all are placeholders or from new tab pages)")
        return

    # Try to load nicer fonts
    try:
        # Try different font options in order of preference
        # ArialUni is a font that comes with Office and can render most non-alphabet characters
        font_options = [
            "Microsoft YaHei",  # 微软雅黑
            "SimHei",  # 黑体
            "SimSun",  # 宋体
            "Noto Sans CJK SC",  # 思源黑体
            "WenQuanYi Micro Hei",  # 文泉驿微米黑
            "Helvetica",
            "Arial",
            "DejaVuSans",
            "Verdana",
        ]
        font_loaded = False

        for font_name in font_options:
            try:
                if platform.system() == "Windows":
                    # Need to specify the abs font path on Windows
                    font_name = os.path.join(CONFIG.WIN_FONT_DIR, font_name + ".ttf")
                regular_font = ImageFont.truetype(font_name, font_size)
                title_font = ImageFont.truetype(font_name, title_font_size)
                ImageFont.truetype(font_name, goal_font_size)
                font_loaded = True
                break
            except OSError:
                continue

        if not font_loaded:
            raise OSError("No preferred fonts found")

    except OSError:
        regular_font = ImageFont.load_default()
        title_font = ImageFont.load_default()

    # Load logo if requested
    logo = None
    if show_logo:
        try:
            logo = Image.open("./static/browser-use.png")
            # Resize logo to be small (e.g., 40px height)
            logo_height = 150
            aspect_ratio = logo.width / logo.height
            logo_width = int(logo_height * aspect_ratio)
            logo = logo.resize((logo_width, logo_height), Image.Resampling.LANCZOS)
        except Exception as e:
            logger.warning(f"Could not load logo: {e}")

    # Create task frame if requested
    if show_task and task:
        # Find the first non-placeholder screenshot for the task frame
        first_real_screenshot = None
        for item in history.history:
            screenshot_b64 = item.state.get_screenshot()
            if screenshot_b64 and screenshot_b64 != PLACEHOLDER_4PX_SCREENSHOT:
                first_real_screenshot = screenshot_b64
                break

        if first_real_screenshot:
            task_frame = _create_task_frame(
                task,
                first_real_screenshot,
                title_font,  # type: ignore
                regular_font,  # type: ignore
                logo,
                line_spacing,
            )
            images.append(task_frame)
        else:
            logger.warning("No real screenshots found for task frame, skipping task frame")

    # Process each history item with its corresponding screenshot
    for i, (item, screenshot) in enumerate(zip(history.history, screenshots), 1):
        if not screenshot:
            continue

        # Skip placeholder screenshots from about:blank pages
        # These are 4x4 white PNGs encoded as a specific base64 string
        if screenshot == PLACEHOLDER_4PX_SCREENSHOT:
            logger.debug(f"Skipping placeholder screenshot from about:blank page at step {i}")
            continue

        # Skip screenshots from new tab pages
        from ..utils import is_new_tab_page

        if is_new_tab_page(item.state.url):
            logger.debug(f"Skipping screenshot from new tab page ({item.state.url}) at step {i}")
            continue

        # Convert base64 screenshot to PIL Image
        img_data = base64.b64decode(screenshot)
        image = Image.open(io.BytesIO(img_data))

        if show_goals and item.model_output:
            image = _add_overlay_to_image(
                image=image,
                step_number=i,
                goal_text=item.model_output.current_state.next_goal,
                regular_font=regular_font,  # type: ignore
                title_font=title_font,  # type: ignore
                margin=margin,
                logo=logo,
            )

        images.append(image)

    if images:
        # Save the GIF
        images[0].save(
            output_path,
            save_all=True,
            append_images=images[1:],
            duration=duration,
            loop=0,
            optimize=False,
        )
        logger.info(f"Created GIF at {output_path}")
    else:
        logger.warning("No images found in history to create GIF")


def _create_task_frame(
    task: str,
    first_screenshot: str,
    title_font: ImageFont.FreeTypeFont,
    regular_font: ImageFont.FreeTypeFont,
    logo: Image.Image | None = None,
    line_spacing: float = 1.5,
) -> Image.Image:
    """Create initial frame showing the task."""
    from PIL import Image, ImageDraw, ImageFont

    img_data = base64.b64decode(first_screenshot)
    template = Image.open(io.BytesIO(img_data))
    image = Image.new("RGB", template.size, (0, 0, 0))
    draw = ImageDraw.Draw(image)

    # Calculate vertical center of image
    center_y = image.height // 2

    # Draw task text with dynamic font size based on task length
    margin = 140  # Increased margin
    max_width = image.width - (2 * margin)

    # Dynamic font size calculation based on task length
    # Start with base font size (regular + 16)
    base_font_size = regular_font.size + 16
    min_font_size = max(regular_font.size - 10, 16)  # Don't go below 16pt

    # Calculate dynamic font size based on text length and complexity
    # Longer texts get progressively smaller fonts
    text_length = len(task)
    if text_length > 200:
        # For very long text, reduce font size logarithmically
        font_size = max(base_font_size - int(10 * (text_length / 200)), min_font_size)
    else:
        font_size = base_font_size

    # Try to create a larger font, but fall back to regular font if it fails
    try:
        larger_font = ImageFont.truetype(regular_font.path, font_size)  # type: ignore
    except (OSError, AttributeError):
        # Fall back to regular font if .path is not available or font loading fails
        larger_font = regular_font

    # Generate wrapped text with the calculated font size
    wrapped_text = _wrap_text(task, larger_font, max_width)

    # Calculate line height with spacing
    line_height = larger_font.size * line_spacing

    # Split text into lines and draw with custom spacing
    lines = wrapped_text.split("\n")
    total_height = line_height * len(lines)

    # Start position for first line
    text_y = center_y - (total_height / 2) + 50  # Shifted down slightly

    for line in lines:
        # Get line width for centering
        line_bbox = draw.textbbox((0, 0), line, font=larger_font)
        text_x = (image.width - (line_bbox[2] - line_bbox[0])) // 2

        draw.text(
            (text_x, text_y),
            line,
            font=larger_font,
            fill=(255, 255, 255),
        )
        text_y += line_height

    # Add logo if provided (top right corner)
    if logo:
        logo_margin = 20
        logo_x = image.width - logo.width - logo_margin
        image.paste(logo, (logo_x, logo_margin), logo if logo.mode == "RGBA" else None)

    return image


def _add_overlay_to_image(
    image: Image.Image,
    step_number: int,
    goal_text: str,
    regular_font: ImageFont.FreeTypeFont,
    title_font: ImageFont.FreeTypeFont,
    margin: int,
    logo: Image.Image | None = None,
    display_step: bool = True,
    text_color: tuple[int, int, int, int] = (255, 255, 255, 255),
    text_box_color: tuple[int, int, int, int] = (0, 0, 0, 255),
) -> Image.Image:
    """Add step number and goal overlay to an image."""

    from PIL import Image, ImageDraw

    goal_text = decode_unicode_escapes_to_utf8(goal_text)
    image = image.convert("RGBA")
    txt_layer = Image.new("RGBA", image.size, (0, 0, 0, 0))
    draw = ImageDraw.Draw(txt_layer)
    if display_step:
        # Add step number (bottom left)
        step_text = str(step_number)
        step_bbox = draw.textbbox((0, 0), step_text, font=title_font)
        step_width = step_bbox[2] - step_bbox[0]
        step_height = step_bbox[3] - step_bbox[1]

        # Position step number in bottom left
        x_step = margin + 10  # Slight additional offset from edge
        y_step = image.height - margin - step_height - 10  # Slight offset from bottom

        # Draw rounded rectangle background for step number
        padding = 20  # Increased padding
        step_bg_bbox = (
            x_step - padding,
            y_step - padding,
            x_step + step_width + padding,
            y_step + step_height + padding,
        )
        draw.rounded_rectangle(
            step_bg_bbox,
            radius=15,  # Add rounded corners
            fill=text_box_color,
        )

        # Draw step number
        draw.text(
            (x_step, y_step),
            step_text,
            font=title_font,
            fill=text_color,
        )

    # Draw goal text (centered, bottom)
    max_width = image.width - (4 * margin)
    wrapped_goal = _wrap_text(goal_text, title_font, max_width)
    goal_bbox = draw.multiline_textbbox((0, 0), wrapped_goal, font=title_font)
    goal_width = goal_bbox[2] - goal_bbox[0]
    goal_height = goal_bbox[3] - goal_bbox[1]

    # Center goal text horizontally, place above step number
    x_goal = (image.width - goal_width) // 2
    y_goal = y_step - goal_height - padding * 4  # More space between step and goal

    # Draw rounded rectangle background for goal
    padding_goal = 25  # Increased padding for goal
    goal_bg_bbox = (
        x_goal - padding_goal,  # Remove extra space for logo
        y_goal - padding_goal,
        x_goal + goal_width + padding_goal,
        y_goal + goal_height + padding_goal,
    )
    draw.rounded_rectangle(
        goal_bg_bbox,
        radius=15,  # Add rounded corners
        fill=text_box_color,
    )

    # Draw goal text
    draw.multiline_text(
        (x_goal, y_goal),
        wrapped_goal,
        font=title_font,
        fill=text_color,
        align="center",
    )

    # Add logo if provided (top right corner)
    if logo:
        logo_layer = Image.new("RGBA", image.size, (0, 0, 0, 0))
        logo_margin = 20
        logo_x = image.width - logo.width - logo_margin
        logo_layer.paste(logo, (logo_x, logo_margin), logo if logo.mode == "RGBA" else None)
        txt_layer = Image.alpha_composite(logo_layer, txt_layer)

    # Composite and convert
    result = Image.alpha_composite(image, txt_layer)
    return result.convert("RGB")


def _wrap_text(text: str, font: ImageFont.FreeTypeFont, max_width: int) -> str:
    """
    Wrap text to fit within a given width.

    Args:
        text: Text to wrap
        font: Font to use for text
        max_width: Maximum width in pixels

    Returns:
        Wrapped text with newlines
    """
    text = decode_unicode_escapes_to_utf8(text)
    words = text.split()
    lines = []
    current_line = []

    for word in words:
        current_line.append(word)
        line = " ".join(current_line)
        bbox = font.getbbox(line)
        if bbox[2] > max_width:
            if len(current_line) == 1:
                lines.append(current_line.pop())
            else:
                current_line.pop()
                lines.append(" ".join(current_line))
                current_line = [word]

    if current_line:
        lines.append(" ".join(current_line))

    return "\n".join(lines)



"""Unified capability registry, discovery, and provider protocol (RFC-0005, RFC-1004)."""

from .models import (
    STATEFUL_TYPES,
    CapabilityDescriptor,
    CapabilityProvider,
    CapabilityQuery,
    CapabilityType,
    DeterminismClass,
    LatencyClass,
    SideEffectClass,
)
from .providers import (
    AgentCapabilityProvider,
    CliAgentCapabilityProvider,
    MCPCapabilityProvider,
    SkillCapabilityProvider,
    ToolCapabilityProvider,
)
from .registry import CapabilityRegistry, RegistryEvent

__all__ = [
    "AgentCapabilityProvider",
    "CapabilityDescriptor",
    "CapabilityProvider",
    "CapabilityQuery",
    "CapabilityRegistry",
    "CapabilityType",
    "CliAgentCapabilityProvider",
    "DeterminismClass",
    "LatencyClass",
    "MCPCapabilityProvider",
    "RegistryEvent",
    "STATEFUL_TYPES",
    "SideEffectClass",
    "SkillCapabilityProvider",
    "ToolCapabilityProvider",
]



"""Adapter for converting existing noesium toolkits to AtomicTool (RFC-2004 §7.1)."""

from __future__ import annotations

import inspect
from typing import Any, Callable, get_type_hints

from noesium.core.toolify.atomic import AtomicTool, ToolSource


def _extract_schema(func: Callable) -> dict[str, Any]:
    """Best-effort JSON Schema extraction from function signature."""
    try:
        hints = get_type_hints(func)
    except Exception:
        hints = {}
    sig = inspect.signature(func)
    properties: dict[str, Any] = {}
    required: list[str] = []
    for name, param in sig.parameters.items():
        if name in ("self", "cls"):
            continue
        prop: dict[str, Any] = {}
        hint = hints.get(name)
        if hint is str or hint == "str":
            prop["type"] = "string"
        elif hint is int or hint == "int":
            prop["type"] = "integer"
        elif hint is float or hint == "float":
            prop["type"] = "number"
        elif hint is bool or hint == "bool":
            prop["type"] = "boolean"
        else:
            prop["type"] = "string"
        properties[name] = prop
        if param.default is inspect.Parameter.empty:
            required.append(name)
    schema: dict[str, Any] = {"type": "object", "properties": properties}
    if required:
        schema["required"] = required
    return schema


class BuiltinAdapter:
    """Converts existing toolkit functions to AtomicTool instances."""

    @staticmethod
    async def from_toolkit(
        toolkit: Any,
        toolkit_name: str,
    ) -> list[AtomicTool]:
        from noesium.core.toolify.base import AsyncBaseToolkit

        if isinstance(toolkit, AsyncBaseToolkit):
            tools_map = await toolkit.get_filtered_tools_map()
        else:
            tools_map = toolkit.get_filtered_tools_map()

        atomic_tools: list[AtomicTool] = []
        for name, func in tools_map.items():
            schema = _extract_schema(func)
            tool = AtomicTool(
                name=name,
                description=func.__doc__ or f"Tool: {name}",
                input_schema=schema,
                source=ToolSource.BUILTIN,
                toolkit_name=toolkit_name,
                tags=[f"source:builtin", f"toolkit:{toolkit_name}"],
            ).bind(func)
            atomic_tools.append(tool)
        return atomic_tools



"""
Gmail API Service for Browser Use
Handles Gmail API authentication, email reading, and 2FA code extraction.
This service provides a clean interface for agents to interact with Gmail.
"""

import base64
import logging
import os
from pathlib import Path
from typing import Any

import aiofiles
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError

from ...config import CONFIG

logger = logging.getLogger(__name__)


class GmailService:
    """
    Gmail API service for email reading.
    Provides functionality to:
    - Authenticate with Gmail API using OAuth2
    - Read recent emails with filtering
    - Return full email content for agent analysis
    """

    # Gmail API scopes
    SCOPES = ["https://www.googleapis.com/auth/gmail.readonly"]

    def __init__(
        self,
        credentials_file: str | None = None,
        token_file: str | None = None,
        config_dir: str | None = None,
        access_token: str | None = None,
    ):
        """
        Initialize Gmail Service
        Args:
            credentials_file: Path to OAuth credentials JSON from Google Cloud Console
            token_file: Path to store/load access tokens
            config_dir: Directory to store config files (defaults to browser-use config directory)
            access_token: Direct access token (skips file-based auth if provided)
        """
        # Set up configuration directory using browser-use's config system
        if config_dir is None:
            self.config_dir = CONFIG.BROWSER_USE_CONFIG_DIR
        else:
            self.config_dir = Path(config_dir).expanduser().resolve()

        # Ensure config directory exists (only if not using direct token)
        if access_token is None:
            self.config_dir.mkdir(parents=True, exist_ok=True)

        # Set up credential paths
        self.credentials_file = credentials_file or self.config_dir / "gmail_credentials.json"
        self.token_file = token_file or self.config_dir / "gmail_token.json"

        # Direct access token support
        self.access_token = access_token

        self.service = None
        self.creds = None
        self._authenticated = False

    def is_authenticated(self) -> bool:
        """Check if Gmail service is authenticated"""
        return self._authenticated and self.service is not None

    async def authenticate(self) -> bool:
        """
        Handle OAuth authentication and token management
        Returns:
            bool: True if authentication successful, False otherwise
        """
        try:
            logger.info("🔐 Authenticating with Gmail API...")

            # Check if using direct access token
            if self.access_token:
                logger.info("🔑 Using provided access token")
                # Create credentials from access token
                self.creds = Credentials(token=self.access_token, scopes=self.SCOPES)
                # Test token validity by building service
                self.service = build("gmail", "v1", credentials=self.creds)
                self._authenticated = True
                logger.info("✅ Gmail API ready with access token!")
                return True

            # Original file-based authentication flow
            # Try to load existing tokens
            if os.path.exists(self.token_file):
                self.creds = Credentials.from_authorized_user_file(str(self.token_file), self.SCOPES)
                logger.debug("📁 Loaded existing tokens")

            # If no valid credentials, run OAuth flow
            if not self.creds or not self.creds.valid:
                if self.creds and self.creds.expired and self.creds.refresh_token:
                    logger.info("🔄 Refreshing expired tokens...")
                    self.creds.refresh(Request())
                else:
                    logger.info("🌐 Starting OAuth flow...")
                    if not os.path.exists(self.credentials_file):
                        logger.error(
                            f"❌ Gmail credentials file not found: {self.credentials_file}\n"
                            "Please download it from Google Cloud Console:\n"
                            "1. Go to https://console.cloud.google.com/\n"
                            "2. APIs & Services > Credentials\n"
                            "3. Download OAuth 2.0 Client JSON\n"
                            f"4. Save as 'gmail_credentials.json' in {self.config_dir}/"
                        )
                        return False

                    flow = InstalledAppFlow.from_client_secrets_file(str(self.credentials_file), self.SCOPES)
                    # Use specific redirect URI to match OAuth credentials
                    self.creds = flow.run_local_server(port=8080, open_browser=True)

                # Save tokens for next time
                async with aiofiles.open(self.token_file, "w") as token:
                    await token.write(self.creds.to_json())
                logger.info(f"💾 Tokens saved to {self.token_file}")

            # Build Gmail service
            self.service = build("gmail", "v1", credentials=self.creds)
            self._authenticated = True
            logger.info("✅ Gmail API ready!")
            return True

        except Exception as e:
            logger.error(f"❌ Gmail authentication failed: {e}")
            return False

    async def get_recent_emails(
        self, max_results: int = 10, query: str = "", time_filter: str = "1h"
    ) -> list[dict[str, Any]]:
        """
        Get recent emails with optional query filter
        Args:
            max_results: Maximum number of emails to fetch
            query: Gmail search query (e.g., 'from:noreply@example.com')
            time_filter: Time filter (e.g., '5m', '1h', '1d')
        Returns:
            List of email dictionaries with parsed content
        """
        if not self.is_authenticated():
            logger.error("❌ Gmail service not authenticated. Call authenticate() first.")
            return []

        try:
            # Add time filter to query if provided
            if time_filter and "newer_than:" not in query:
                query = f"newer_than:{time_filter} {query}".strip()

            logger.info(f"📧 Fetching {max_results} recent emails...")
            if query:
                logger.debug(f"🔍 Query: {query}")

            # Get message list
            assert self.service is not None
            results = self.service.users().messages().list(userId="me", maxResults=max_results, q=query).execute()

            messages = results.get("messages", [])
            if not messages:
                logger.info("📭 No messages found")
                return []

            logger.info(f"📨 Found {len(messages)} messages, fetching details...")

            # Get full message details
            emails = []
            for i, message in enumerate(messages, 1):
                logger.debug(f"📖 Reading email {i}/{len(messages)}...")

                full_message = (
                    self.service.users().messages().get(userId="me", id=message["id"], format="full").execute()
                )

                email_data = self._parse_email(full_message)
                emails.append(email_data)

            return emails

        except HttpError as error:
            logger.error(f"❌ Gmail API error: {error}")
            return []
        except Exception as e:
            logger.error(f"❌ Unexpected error fetching emails: {e}")
            return []

    def _parse_email(self, message: dict[str, Any]) -> dict[str, Any]:
        """Parse Gmail message into readable format"""
        headers = {h["name"]: h["value"] for h in message["payload"]["headers"]}

        return {
            "id": message["id"],
            "thread_id": message["threadId"],
            "subject": headers.get("Subject", ""),
            "from": headers.get("From", ""),
            "to": headers.get("To", ""),
            "date": headers.get("Date", ""),
            "timestamp": int(message["internalDate"]),
            "body": self._extract_body(message["payload"]),
            "raw_message": message,
        }

    def _extract_body(self, payload: dict[str, Any]) -> str:
        """Extract email body from payload"""
        body = ""

        if payload.get("body", {}).get("data"):
            # Simple email body
            body = base64.urlsafe_b64decode(payload["body"]["data"]).decode("utf-8")
        elif payload.get("parts"):
            # Multi-part email
            for part in payload["parts"]:
                if part["mimeType"] == "text/plain" and part.get("body", {}).get("data"):
                    part_body = base64.urlsafe_b64decode(part["body"]["data"]).decode("utf-8")
                    body += part_body
                elif part["mimeType"] == "text/html" and not body and part.get("body", {}).get("data"):
                    # Fallback to HTML if no plain text
                    body = base64.urlsafe_b64decode(part["body"]["data"]).decode("utf-8")

        return body



from typing import List, Optional

from noesium.core.llm import BaseLLMClient
from noesium.core.utils.logging import get_logger

from .models import AskuraConfig, AskuraState
from .reflection import Reflection

logger = get_logger(__name__)


class Summarizer:
    def __init__(
        self, config: AskuraConfig, llm_client: Optional[BaseLLMClient] = None, reflection: Optional[Reflection] = None
    ):
        self.config = config
        self.llm = llm_client
        self.reflection = reflection

    def is_ready_to_summarize(self, state: AskuraState) -> bool:
        # Summarize only when all required slots are complete
        return len(self.reflection.missing_slots(state)) == 0 and state.turns > 1

    def summarize(self, state: AskuraState) -> str:
        information_slots = state.extracted_info
        summary_parts: List[str] = []
        for slot in self.config.information_slots:
            if information_slots.get(slot.name):
                summary_parts.append(f"{slot.name}: {information_slots[slot.name]}")
        return "Summary: " + " | ".join(summary_parts) if summary_parts else "Conversation completed."



"""Skill composition model (RFC-2004 §9)."""

from __future__ import annotations

from typing import Any, Callable

from pydantic import BaseModel, Field, PrivateAttr
from uuid_extensions import uuid7str

from noesium.core.exceptions import SkillNotFoundError


class Skill(BaseModel):
    """A named composition of one or more AtomicTools."""

    skill_id: str = Field(default_factory=lambda: uuid7str())
    name: str
    description: str
    tool_names: list[str]
    input_schema: dict[str, Any] = Field(default_factory=dict)
    output_schema: dict[str, Any] | None = None
    tags: list[str] = Field(default_factory=list)

    _executor_fn: Callable | None = PrivateAttr(default=None)

    def bind_executor(self, fn: Callable) -> Skill:
        self._executor_fn = fn
        return self

    async def execute(
        self,
        tool_registry: Any,
        tool_executor: Any,
        context: Any,
        **kwargs: Any,
    ) -> Any:
        if self._executor_fn:
            return await self._executor_fn(
                tool_registry=tool_registry,
                tool_executor=tool_executor,
                context=context,
                **kwargs,
            )
        raise NotImplementedError("Skill has no bound executor")


class SkillRegistry:
    def __init__(self) -> None:
        self._skills: dict[str, Skill] = {}

    def register(self, skill: Skill) -> None:
        self._skills[skill.name] = skill

    def get(self, name: str) -> Skill:
        if name not in self._skills:
            raise SkillNotFoundError(f"Skill '{name}' not registered")
        return self._skills[name]

    def list_skills(self) -> list[Skill]:
        return list(self._skills.values())



from typing import Any, Dict

from .models import AskuraState


class Memory:
    def __init__(self):
        pass

    def save(self, state: AskuraState):
        pass

    def load(self, session_id: str) -> Dict[str, Any]:
        return {}



"""Tool source adapters (RFC-2004 §7)."""

from .builtin_adapter import BuiltinAdapter
from .function_adapter import FunctionAdapter
from .langchain_adapter import LangChainAdapter
from .mcp_adapter import MCPAdapter

__all__ = [
    "BuiltinAdapter",
    "FunctionAdapter",
    "LangChainAdapter",
    "MCPAdapter",
]



"""Code-use agent service - Jupyter notebook-like code execution for browser automation."""

import asyncio
import datetime
import html
import json
import logging
import re
import traceback
from pathlib import Path
from typing import Any

from uuid_extensions import uuid7str

from noesium.subagents.bu.adapters.llm_adapter import (
    AssistantMessage,
    BaseChatModel,
    BaseMessage,
    ContentPartImageParam,
    ContentPartTextParam,
    ImageURL,
    UserMessage,
)
from noesium.subagents.bu.browser import BrowserSession
from noesium.subagents.bu.browser.profile import BrowserProfile
from noesium.subagents.bu.dom.service import DomService
from noesium.subagents.bu.filesystem.file_system import FileSystem
from noesium.subagents.bu.screenshots.service import ScreenshotService
from noesium.subagents.bu.tokens.service import TokenCost
from noesium.subagents.bu.tokens.views import UsageSummary
from noesium.subagents.bu.tools.service import CodeAgentTools, Tools
from noesium.subagents.bu.utils import get_browser_use_version

from .formatting import format_browser_state_for_llm
from .namespace import EvaluateError, create_namespace
from .utils import detect_token_limit_issue, extract_code_blocks, extract_url_from_task, truncate_message_content
from .views import (
    CellType,
    CodeAgentHistory,
    CodeAgentHistoryList,
    CodeAgentModelOutput,
    CodeAgentResult,
    CodeAgentState,
    CodeAgentStepMetadata,
    ExecutionStatus,
    NotebookSession,
)

logger = logging.getLogger(__name__)


class CodeAgent:
    """
    Agent that executes Python code in a notebook-like environment for browser automation.

    This agent provides a Jupyter notebook-like interface where the LLM writes Python code
    that gets executed in a persistent namespace with browser control functions available.
    """

    def __init__(
        self,
        task: str,
        # Optional parameters
        llm: BaseChatModel | None = None,
        browser_session: BrowserSession | None = None,
        browser: BrowserSession | None = None,  # Alias for browser_session
        tools: Tools | None = None,
        controller: Tools | None = None,  # Alias for tools
        # Agent settings
        page_extraction_llm: BaseChatModel | None = None,
        file_system: FileSystem | None = None,
        available_file_paths: list[str] | None = None,
        sensitive_data: dict[str, str | dict[str, str]] | None = None,
        max_steps: int = 100,
        max_failures: int = 8,
        max_validations: int = 0,
        use_vision: bool = True,
        calculate_cost: bool = False,
        demo_mode: bool | None = None,
        **kwargs,
    ):
        """
        Initialize the code-use agent.

        Args:
                task: The task description for the agent
                browser_session: Optional browser session (will be created if not provided) [DEPRECATED: use browser]
                browser: Optional browser session (cleaner API)
                tools: Optional Tools instance (will create default if not provided)
                controller: Optional Tools instance
                page_extraction_llm: Optional LLM for page extraction
                file_system: Optional file system for file operations
                available_file_paths: Optional list of available file paths
                sensitive_data: Optional sensitive data dictionary
                max_steps: Maximum number of execution steps
                max_failures: Maximum consecutive errors before termination (default: 8)
                max_validations: Maximum number of times to run the validator agent (default: 0)
                use_vision: Whether to include screenshots in LLM messages (default: True)
                calculate_cost: Whether to calculate token costs (default: False)
                demo_mode: Enable the in-browser demo panel for live logging (default: False)
                llm: Optional ChatBrowserUse LLM instance (will create default if not provided)
                **kwargs: Additional keyword arguments for compatibility (ignored)
        """
        # Log and ignore unknown kwargs for compatibility
        if kwargs:
            logger.debug(f"Ignoring additional kwargs for CodeAgent compatibility: {list(kwargs.keys())}")

        if llm is None:
            # For noesium, we require an LLM to be provided
            raise RuntimeError("CodeAgent requires an LLM to be provided. Please pass an llm parameter.")

        # No longer restrict to ChatBrowserUse - work with any LLM adapter

        # Handle browser vs browser_session parameter (browser takes precedence)
        if browser and browser_session:
            raise ValueError(
                'Cannot specify both "browser" and "browser_session" parameters. Use "browser" for the cleaner API.'
            )
        browser_session = browser or browser_session

        # Handle controller vs tools parameter (controller takes precedence)
        if controller and tools:
            raise ValueError(
                'Cannot specify both "controller" and "tools" parameters. Use "controller" for the cleaner API.'
            )
        tools = controller or tools

        # Store browser_profile for creating browser session if needed
        self._demo_mode_enabled = False
        if browser_session is None:
            profile_kwargs: dict[str, Any] = {}
            if demo_mode is not None:
                profile_kwargs["demo_mode"] = demo_mode
            self._browser_profile_for_init = BrowserProfile(**profile_kwargs)
        else:
            self._browser_profile_for_init = None

        self.task = task
        self.llm = llm
        self.browser_session = browser_session
        if self.browser_session:
            if demo_mode is not None and self.browser_session.browser_profile.demo_mode != demo_mode:
                self.browser_session.browser_profile = self.browser_session.browser_profile.model_copy(
                    update={"demo_mode": demo_mode}
                )
            self._demo_mode_enabled = bool(self.browser_session.browser_profile.demo_mode)
        self.tools = tools or CodeAgentTools()
        self.page_extraction_llm = page_extraction_llm
        self.file_system = file_system if file_system is not None else FileSystem(base_dir="./")
        self.available_file_paths = available_file_paths or []
        self.sensitive_data = sensitive_data
        self.max_steps = max_steps
        self.max_failures = max_failures
        self.max_validations = max_validations
        self.use_vision = use_vision

        self.session = NotebookSession()
        self.namespace: dict[str, Any] = {}
        self._llm_messages: list[BaseMessage] = []  # Internal LLM conversation history
        self.complete_history: list[CodeAgentHistory] = []  # Type-safe history with model_output and result
        self.dom_service: DomService | None = None
        self._last_browser_state_text: str | None = None  # Track last browser state text
        self._last_screenshot: str | None = None  # Track last screenshot (base64)
        self._consecutive_errors = 0  # Track consecutive errors for auto-termination
        self._validation_count = 0  # Track number of validator runs
        self._last_llm_usage: Any | None = None  # Track last LLM call usage stats
        self._step_start_time = 0.0  # Track step start time for duration calculation
        self.usage_summary: UsageSummary | None = None  # Track usage summary across run for history property
        self._sample_output_added = False  # Track whether preview cell already created

        # Initialize screenshot service for eval tracking
        self.id = uuid7str()
        timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        base_tmp = Path("/tmp")
        self.agent_directory = base_tmp / f"browser_use_code_agent_{self.id}_{timestamp}"
        self.screenshot_service = ScreenshotService(agent_directory=self.agent_directory)

        # Initialize token cost service for usage tracking
        self.token_cost_service = TokenCost(include_cost=calculate_cost)
        self.token_cost_service.register_llm(llm)
        if page_extraction_llm:
            self.token_cost_service.register_llm(page_extraction_llm)

        # Set version and source for telemetry
        self.version = get_browser_use_version()
        try:
            package_root = Path(__file__).parent.parent.parent
            repo_files = [".git", "README.md", "docs", "examples"]
            if all(Path(package_root / file).exists() for file in repo_files):
                self.source = "git"
            else:
                self.source = "pip"
        except Exception:
            self.source = "unknown"

    async def run(self, max_steps: int | None = None) -> NotebookSession:
        """
        Run the agent to complete the task.

        Args:
                max_steps: Optional override for maximum number of steps (uses __init__ value if not provided)

        Returns:
                The notebook session with all executed cells
        """
        # Use override if provided, otherwise use value from __init__
        steps_to_run = max_steps if max_steps is not None else self.max_steps
        self.max_steps = steps_to_run
        # Start browser if not provided
        if self.browser_session is None:
            assert self._browser_profile_for_init is not None
            self.browser_session = BrowserSession(browser_profile=self._browser_profile_for_init)
            await self.browser_session.start()

        if self.browser_session:
            self._demo_mode_enabled = bool(self.browser_session.browser_profile.demo_mode)
            if self._demo_mode_enabled and getattr(self.browser_session.browser_profile, "headless", False):
                logger.warning(
                    "Demo mode is enabled but the browser is headless=True; set headless=False to view the panel."
                )
            if self._demo_mode_enabled:
                await self._demo_mode_log(f"Started CodeAgent task: {self.task}", "info", {"tag": "task"})

        # Initialize DOM service with cross-origin iframe support enabled
        self.dom_service = DomService(
            browser_session=self.browser_session,
            cross_origin_iframes=True,  # Enable for code-use agent to access forms in iframes
        )

        # Create namespace with all tools
        self.namespace = create_namespace(
            browser_session=self.browser_session,
            tools=self.tools,
            page_extraction_llm=self.page_extraction_llm,
            file_system=self.file_system,
            available_file_paths=self.available_file_paths,
            sensitive_data=self.sensitive_data,
        )

        # Initialize conversation with task
        self._llm_messages.append(UserMessage(content=f"Task: {self.task}"))

        # Track agent run error for telemetry
        agent_run_error: str | None = None
        should_delay_close = False

        # Extract URL from task and navigate if found
        initial_url = extract_url_from_task(self.task)
        if initial_url:
            try:
                logger.info(f"Extracted URL from task, navigating to: {initial_url}")
                # Use the navigate action from namespace
                await self.namespace["navigate"](initial_url)
                # Wait for page load
                await asyncio.sleep(2)

                # Record this navigation as a cell in the notebook
                nav_code = f"await navigate('{initial_url}')"
                cell = self.session.add_cell(source=nav_code)
                cell.status = ExecutionStatus.SUCCESS
                cell.execution_count = self.session.increment_execution_count()
                cell.output = f"Navigated to {initial_url}"

                # Get browser state after navigation for the cell
                if self.dom_service:
                    try:
                        browser_state_text, _ = await self._get_browser_state()
                        cell.browser_state = browser_state_text
                    except Exception as state_error:
                        logger.debug(f"Failed to capture browser state for initial navigation cell: {state_error}")

            except Exception as e:
                logger.warning(f"Failed to navigate to extracted URL {initial_url}: {e}")
                # Record failed navigation as error cell
                nav_code = f"await navigate('{initial_url}')"
                cell = self.session.add_cell(source=nav_code)
                cell.status = ExecutionStatus.ERROR
                cell.execution_count = self.session.increment_execution_count()
                cell.error = str(e)

        # Get initial browser state before first LLM call
        if self.browser_session and self.dom_service:
            try:
                browser_state_text, screenshot = await self._get_browser_state()
                self._last_browser_state_text = browser_state_text
                self._last_screenshot = screenshot
            except Exception as e:
                logger.warning(f"Failed to get initial browser state: {e}")

        # Main execution loop
        for step in range(self.max_steps):
            logger.info(f"\n\n\n\n\n\n\nStep {step + 1}/{self.max_steps}")
            await self._demo_mode_log(f"Starting step {step + 1}/{self.max_steps}", "info", {"step": step + 1})

            # Start timing this step
            self._step_start_time = datetime.datetime.now().timestamp()

            # Check if we're approaching the step limit or error limit and inject warning
            steps_remaining = self.max_steps - step - 1
            errors_remaining = self.max_failures - self._consecutive_errors

            should_warn = (
                steps_remaining <= 1  # Last step or next to last
                or errors_remaining <= 1  # One more error will terminate
                or (steps_remaining <= 2 and self._consecutive_errors >= 2)  # Close to both limits
            )

            if should_warn:
                warning_message = (
                    f"\n\n⚠️ CRITICAL WARNING: You are approaching execution limits!\n"
                    f"- Steps remaining: {steps_remaining + 1}\n"
                    f"- Consecutive errors: {self._consecutive_errors}/{self.max_failures}\n\n"
                    f"YOU MUST call done() in your NEXT response, even if the task is incomplete:\n"
                    f"- Set success=False if you couldn't complete the task\n"
                    f"- Return EVERYTHING you found so far (partial data is better than nothing)\n"
                    f"- Include any variables you've stored (products, all_data, etc.)\n"
                    f"- Explain what worked and what didn't\n\n"
                    f"Without done(), the user will receive NOTHING."
                )
                self._llm_messages.append(UserMessage(content=warning_message))

            try:
                # Fetch fresh browser state right before LLM call (only if not already set)
                if not self._last_browser_state_text and self.browser_session and self.dom_service:
                    try:
                        logger.debug("🔍 Fetching browser state before LLM call...")
                        browser_state_text, screenshot = await self._get_browser_state()
                        self._last_browser_state_text = browser_state_text
                        self._last_screenshot = screenshot

                        # # Log browser state
                        # if len(browser_state_text) > 2000:
                        # 	logger.info(
                        # 		f'Browser state (before LLM):\n{browser_state_text[:2000]}...\n[Truncated, full state {len(browser_state_text)} chars sent to LLM]'
                        # 	)
                        # else:
                        # 	logger.info(f'Browser state (before LLM):\n{browser_state_text}')
                    except Exception as e:
                        logger.warning(f"Failed to get browser state before LLM call: {e}")

                # Get code from LLM (this also adds to self._llm_messages)
                try:
                    code, full_llm_response = await self._get_code_from_llm(step_number=step + 1)
                except Exception as llm_error:
                    # LLM call failed - count as consecutive error and retry
                    self._consecutive_errors += 1
                    logger.warning(
                        f"LLM call failed (consecutive errors: {self._consecutive_errors}/{self.max_failures}), retrying: {llm_error}"
                    )
                    await self._demo_mode_log(
                        f"LLM call failed: {llm_error}",
                        "error",
                        {"step": step + 1},
                    )

                    # Check if we've hit the consecutive error limit
                    if self._consecutive_errors >= self.max_failures:
                        logger.error(f"Terminating: {self.max_failures} consecutive LLM failures")
                        break

                    await asyncio.sleep(1)  # Brief pause before retry
                    continue

                if not code or code.strip() == "":
                    # If task is already done, empty code is fine (LLM explaining completion)
                    if self._is_task_done():
                        logger.info("Task already marked as done, LLM provided explanation without code")
                        # Add the text response to history as a non-code step
                        await self._add_step_to_complete_history(
                            model_output_code="",
                            full_llm_response=full_llm_response,
                            output=full_llm_response,  # Treat the explanation as output
                            error=None,
                            screenshot_path=await self._capture_screenshot(step + 1),
                        )
                        break  # Exit the loop since task is done

                    logger.warning("LLM returned empty code")
                    self._consecutive_errors += 1

                    # new state
                    if self.browser_session and self.dom_service:
                        try:
                            browser_state_text, screenshot = await self._get_browser_state()
                            self._last_browser_state_text = browser_state_text
                            self._last_screenshot = screenshot
                        except Exception as e:
                            logger.warning(f"Failed to get new browser state: {e}")
                    continue

                # Execute code blocks sequentially if multiple python blocks exist
                # This allows JS/bash blocks to be injected into namespace before Python code uses them
                all_blocks = self.namespace.get("_all_code_blocks", {})
                python_blocks = [k for k in sorted(all_blocks.keys()) if k.startswith("python_")]

                if len(python_blocks) > 1:
                    # Multiple Python blocks - execute each sequentially
                    output = None
                    error = None

                    for i, block_key in enumerate(python_blocks):
                        logger.info(f"Executing Python block {i + 1}/{len(python_blocks)}")
                        block_code = all_blocks[block_key]
                        block_output, block_error, _ = await self._execute_code(block_code)

                        # Accumulate outputs
                        if block_output:
                            output = (output or "") + block_output
                        if block_error:
                            error = block_error
                            # Stop on first error
                            break
                else:
                    # Single Python block - execute normally
                    output, error, _ = await self._execute_code(code)

                # Track consecutive errors
                if error:
                    self._consecutive_errors += 1
                    logger.warning(f"Consecutive errors: {self._consecutive_errors}/{self.max_failures}")

                    # Check if we've hit the consecutive error limit
                    if self._consecutive_errors >= self.max_failures:
                        logger.error(
                            f"Terminating: {self.max_failures} consecutive errors reached. The agent is unable to make progress."
                        )
                        await self._demo_mode_log(
                            f"Terminating after {self.max_failures} consecutive errors without progress.",
                            "error",
                            {"step": step + 1},
                        )
                        # Add termination message to complete history before breaking
                        await self._add_step_to_complete_history(
                            model_output_code=code,
                            full_llm_response=f"[Terminated after {self.max_failures} consecutive errors]",
                            output=None,
                            error=f"Auto-terminated: {self.max_failures} consecutive errors without progress",
                            screenshot_path=None,
                        )
                        break
                else:
                    # Reset consecutive error counter on success
                    self._consecutive_errors = 0

                # Check if task is done - validate completion first if not at limits
                if self._is_task_done():
                    # Get the final result from namespace (from done() call)
                    final_result: str | None = self.namespace.get("_task_result")  # type: ignore[assignment]

                    # Check if we should validate (not at step/error limits and under max validations)
                    steps_remaining = self.max_steps - step - 1
                    should_validate = (
                        self._validation_count < self.max_validations  # Haven't exceeded max validations
                        and steps_remaining >= 4  # At least 4 steps away from limit
                        and self._consecutive_errors < 3  # Not close to error limit (8 consecutive)
                    )

                    if should_validate:
                        self._validation_count += 1
                        logger.info("Validating task completion with LLM...")
                        from .namespace import validate_task_completion

                        is_complete, reasoning = await validate_task_completion(
                            task=self.task,
                            output=final_result,
                            llm=self.llm,
                        )

                        if not is_complete:
                            # Task not truly complete - inject feedback and continue
                            logger.warning("Validator: Task not complete, continuing...")
                            validation_feedback = (
                                f"\n\n⚠️ VALIDATOR FEEDBACK:\n"
                                f"Your done() call was rejected. The task is NOT complete yet.\n\n"
                                f"Validation reasoning:\n{reasoning}\n\n"
                                f"You must continue working on the task. Analyze what is missing and complete it.\n"
                                f"Do NOT call done() again until the task is truly finished."
                            )

                            # Clear the done flag so execution continues
                            self.namespace["_task_done"] = False
                            self.namespace.pop("_task_result", None)
                            self.namespace.pop("_task_success", None)

                            # Add validation feedback to LLM messages
                            self._llm_messages.append(UserMessage(content=validation_feedback))

                            # Don't override output - let execution continue normally
                        else:
                            logger.info("Validator: Task complete")
                            # Override output with done message for final step
                            if final_result:
                                output = final_result
                    else:
                        # At limits - skip validation and accept done()
                        if self._validation_count >= self.max_validations:
                            logger.info(
                                f"Reached max validations ({self.max_validations}) - skipping validation and accepting done()"
                            )
                        else:
                            logger.info("At step/error limits - skipping validation")
                        if final_result:
                            output = final_result

                if output:
                    # Check if this is the final done() output
                    if self._is_task_done():
                        # Show done() output more prominently
                        logger.info(
                            f"✓ Task completed - Final output from done():\n{output[:300] if len(output) > 300 else output}"
                        )
                        # Also show files_to_display if they exist in namespace
                        attachments: list[str] | None = self.namespace.get("_task_attachments")  # type: ignore[assignment]
                        if attachments:
                            logger.info(f'Files displayed: {", ".join(attachments)}')
                    else:
                        logger.info(f"Code output:\n{output}")

                # Browser state is now only logged when fetched before LLM call (not after execution)

                # Take screenshot for eval tracking
                screenshot_path = await self._capture_screenshot(step + 1)

                # Add step to complete_history for eval system
                await self._add_step_to_complete_history(
                    model_output_code=code,
                    full_llm_response=full_llm_response,
                    output=output,
                    error=error,
                    screenshot_path=screenshot_path,
                )

                # Check if task is done (after validation)
                if self._is_task_done():
                    # Get the final result from namespace
                    final_result: str | None = self.namespace.get("_task_result", output)  # type: ignore[assignment]
                    logger.info("Task completed successfully")
                    if final_result:
                        logger.info(f"Final result: {final_result}")
                        self._add_sample_output_cell(final_result)
                    if self._demo_mode_enabled:
                        await self._demo_mode_log(
                            f'Final Result: {final_result or "Task completed"}',
                            "success",
                            {"tag": "task"},
                        )
                    should_delay_close = True
                    break
                # If validation rejected done(), continue to next iteration
                # The feedback message has already been added to _llm_messages

                # Add result to LLM messages for next iteration (without browser state)
                result_message = self._format_execution_result(code, output, error, current_step=step + 1)
                truncated_result = truncate_message_content(result_message)
                self._llm_messages.append(UserMessage(content=truncated_result))

            except Exception as e:
                logger.error(f"Error in step {step + 1}: {e}")
                traceback.print_exc()
                break
        else:
            # Loop completed without break - max_steps reached
            logger.warning(f"Maximum steps ({self.max_steps}) reached without task completion")
            await self._demo_mode_log(
                f"Maximum steps ({self.max_steps}) reached without completing the task.",
                "error",
                {"tag": "task"},
            )

        # If task is not done, capture the last step's output as partial result
        if not self._is_task_done() and self.complete_history:
            # Get the last step's output/error and use it as final extracted_content
            last_step = self.complete_history[-1]
            last_result = last_step.result[0] if last_step.result else None
            last_output = last_result.extracted_content if last_result else None
            last_error = last_result.error if last_result else None

            # Build a partial result message from the last step
            partial_result_parts = []
            partial_result_parts.append(f"Task incomplete - reached step limit ({self.max_steps} steps).")
            partial_result_parts.append("Last step output:")

            if last_output:
                partial_result_parts.append(f"\nOutput: {last_output}")
            if last_error:
                partial_result_parts.append(f"\nError: {last_error}")

            # Add any accumulated variables that might contain useful data
            data_vars = []
            for var_name in sorted(self.namespace.keys()):
                if not var_name.startswith("_") and var_name not in {
                    "json",
                    "asyncio",
                    "csv",
                    "re",
                    "datetime",
                    "Path",
                }:
                    var_value = self.namespace[var_name]
                    # Check if it's a list or dict that might contain collected data
                    if isinstance(var_value, (list, dict)) and var_value:
                        data_vars.append(f"  - {var_name}: {type(var_value).__name__} with {len(var_value)} items")

            if data_vars:
                partial_result_parts.append("\nVariables in namespace that may contain partial data:")
                partial_result_parts.extend(data_vars)

            partial_result = "\n".join(partial_result_parts)

            # Update the last step's extracted_content with this partial result
            if last_result:
                last_result.extracted_content = partial_result
                last_result.is_done = False
                last_result.success = False

            logger.info(f"\nPartial result captured from last step:\n{partial_result}")
            if self._demo_mode_enabled:
                await self._demo_mode_log(f"Partial result:\n{partial_result}", "error", {"tag": "task"})

        # Log final summary if task was completed
        if self._is_task_done():
            logger.info("\n" + "=" * 60)
            logger.info("TASK COMPLETED SUCCESSFULLY")
            logger.info("=" * 60)
            final_result: str | None = self.namespace.get("_task_result")  # type: ignore[assignment]
            if final_result:
                logger.info(f"\nFinal Output:\n{final_result}")
                self._add_sample_output_cell(final_result)

            attachments: list[str] | None = self.namespace.get("_task_attachments")  # type: ignore[assignment]
            if attachments:
                logger.info(f"\nFiles Attached:\n{chr(10).join(attachments)}")
            logger.info("=" * 60 + "\n")
            if self._demo_mode_enabled and not should_delay_close:
                await self._demo_mode_log(
                    f'Final Result: {final_result or "Task completed"}',
                    "success",
                    {"tag": "task"},
                )
                should_delay_close = True

        # Auto-close browser if keep_alive is False
        if should_delay_close and self._demo_mode_enabled:
            await asyncio.sleep(30)
        await self.close()

        # Store usage summary for history property
        self.usage_summary = await self.token_cost_service.get_usage_summary()

        # Log token usage summary
        await self.token_cost_service.log_usage_summary()

        # Store history data in session for history property
        self.session._complete_history = self.complete_history
        self.session._usage_summary = self.usage_summary

        return self.session

    async def _get_code_from_llm(self, step_number: int | None = None) -> tuple[str, str]:
        """Get Python code from the LLM.

        Returns:
                Tuple of (extracted_code, full_llm_response)
        """
        # Prepare messages for this request
        # Include browser state as separate message if available (not accumulated in history)
        messages_to_send = self._llm_messages.copy()

        if self._last_browser_state_text:
            # Create message with optional screenshot
            if self.use_vision and self._last_screenshot:
                # Build content with text + screenshot
                content_parts: list[ContentPartTextParam | ContentPartImageParam] = [
                    ContentPartTextParam(text=self._last_browser_state_text)
                ]

                # Add screenshot
                content_parts.append(
                    ContentPartImageParam(
                        image_url=ImageURL(
                            url=f"data:image/png;base64,{self._last_screenshot}",
                            media_type="image/png",
                            detail="auto",
                        ),
                    )
                )

                messages_to_send.append(UserMessage(content=content_parts))
            else:
                # Text only
                messages_to_send.append(UserMessage(content=self._last_browser_state_text))

            # Clear browser state after including it so it's only in this request
            self._last_browser_state_text = None
            self._last_screenshot = None

        # Call LLM with message history (including temporary browser state message)
        response = await self.llm.ainvoke(messages_to_send)

        # Store usage stats from this LLM call
        self._last_llm_usage = response.usage

        # Log the LLM's raw output for debugging
        logger.info(f"LLM Response:\n{response.completion}")
        await self._demo_mode_log(
            f"LLM Response:\n{response.completion}",
            "thought",
            {"step": step_number} if step_number else None,
        )

        # Check for token limit or repetition issues
        max_tokens = getattr(self.llm, "max_tokens", None)
        completion_tokens = response.usage.completion_tokens if response.usage else None
        is_problematic, issue_message = detect_token_limit_issue(
            completion=response.completion,
            completion_tokens=completion_tokens,
            max_tokens=max_tokens,
            stop_reason=response.stop_reason,
        )

        if is_problematic:
            logger.warning(f"Token limit issue detected: {issue_message}")
            # Don't add the bad response to history
            # Instead, inject a system message prompting recovery
            recovery_prompt = (
                f"Your previous response hit a token limit or became repetitive: {issue_message}\n\n"
                "Please write a SHORT plan (2 sentences) for what to do next, then execute ONE simple action."
            )
            self._llm_messages.append(UserMessage(content=recovery_prompt))
            # Return a controlled error message instead of corrupted code
            return "", f"[Token limit error: {issue_message}]"

        # Store the full response
        full_response = response.completion

        # Extract code blocks from response
        # Support multiple code block types: python, js, bash, markdown
        code_blocks = extract_code_blocks(response.completion)

        # Inject non-python blocks into namespace as variables
        # Track which variables are code blocks for browser state display
        if "_code_block_vars" not in self.namespace:
            self.namespace["_code_block_vars"] = set()

        for block_type, block_content in code_blocks.items():
            if not block_type.startswith("python"):
                # Store js, bash, markdown blocks (and named variants) as variables in namespace
                self.namespace[block_type] = block_content
                self.namespace["_code_block_vars"].add(block_type)
                print(f"→ Code block variable: {block_type} (str, {len(block_content)} chars)")
                logger.debug(f"Injected {block_type} block into namespace ({len(block_content)} chars)")

        # Store all code blocks for sequential execution
        self.namespace["_all_code_blocks"] = code_blocks

        # Get Python code if it exists
        # If no python block exists and no other code blocks exist, return empty string to skip execution
        # This prevents treating plain text explanations as code
        code = code_blocks.get("python", response.completion)

        # Add to LLM messages (truncate for history to save context)
        truncated_completion = truncate_message_content(response.completion)
        self._llm_messages.append(AssistantMessage(content=truncated_completion))

        return code, full_response

    def _print_variable_info(self, var_name: str, value: Any) -> None:
        """Print compact info about a variable assignment."""
        # Skip built-in modules and known imports
        skip_names = {
            "json",
            "asyncio",
            "csv",
            "re",
            "datetime",
            "Path",
            "pd",
            "np",
            "plt",
            "requests",
            "BeautifulSoup",
            "PdfReader",
            "browser",
            "file_system",
        }
        if var_name in skip_names:
            return

        # Skip code block variables (already printed)
        if "_code_block_vars" in self.namespace and var_name in self.namespace.get("_code_block_vars", set()):
            return

        # Print compact variable info
        if isinstance(value, (list, dict)):
            preview = str(value)[:100]
            print(f"→ Variable: {var_name} ({type(value).__name__}, len={len(value)}, preview={preview}...)")
        elif isinstance(value, str) and len(value) > 50:
            print(f"→ Variable: {var_name} (str, {len(value)} chars, preview={value[:50]}...)")
        elif callable(value):
            print(f"→ Variable: {var_name} (function)")
        else:
            print(f"→ Variable: {var_name} ({type(value).__name__}, value={repr(value)[:50]})")

    async def _execute_code(self, code: str) -> tuple[str | None, str | None, str | None]:
        """
        Execute Python code in the namespace.

        Args:
                code: The Python code to execute

        Returns:
                Tuple of (output, error, browser_state)
        """
        # Create new cell
        cell = self.session.add_cell(source=code)
        cell.status = ExecutionStatus.RUNNING
        cell.execution_count = self.session.increment_execution_count()

        output = None
        error = None

        try:
            # Capture output
            import ast
            import io
            import sys

            old_stdout = sys.stdout
            sys.stdout = io.StringIO()

            try:
                # Add asyncio to namespace if not already there
                if "asyncio" not in self.namespace:
                    self.namespace["asyncio"] = asyncio

                # Store the current code in namespace for done() validation
                self.namespace["_current_cell_code"] = code
                # Store consecutive errors count for done() validation
                self.namespace["_consecutive_errors"] = self._consecutive_errors

                # Check if code contains await expressions - if so, wrap in async function
                # This mimics how Jupyter/IPython handles top-level await
                try:
                    tree = ast.parse(code, mode="exec")
                    has_await = any(
                        isinstance(node, (ast.Await, ast.AsyncWith, ast.AsyncFor)) for node in ast.walk(tree)
                    )
                except SyntaxError:
                    # If parse fails, let exec handle the error
                    has_await = False

                if has_await:
                    # When code has await, we must wrap in async function
                    # To make variables persist naturally (like Jupyter without needing 'global'):
                    # 1. Extract all assigned variable names from the code
                    # 2. Inject 'global' declarations for variables that already exist in namespace
                    # 3. Extract user's explicit global declarations and pre-define those vars
                    # 4. Return locals() so we can update namespace with new variables

                    # Find all variable names being assigned + user's explicit globals
                    try:
                        assigned_names = set()
                        user_global_names = set()

                        for node in ast.walk(tree):
                            if isinstance(node, ast.Assign):
                                for target in node.targets:
                                    if isinstance(target, ast.Name):
                                        assigned_names.add(target.id)
                            elif isinstance(node, ast.AugAssign) and isinstance(node.target, ast.Name):
                                assigned_names.add(node.target.id)
                            elif isinstance(node, (ast.AnnAssign, ast.NamedExpr)):
                                if hasattr(node, "target") and isinstance(node.target, ast.Name):
                                    assigned_names.add(node.target.id)
                            elif isinstance(node, ast.Global):
                                # Track user's explicit global declarations
                                user_global_names.update(node.names)

                        # Pre-define any user-declared globals that don't exist yet
                        # This prevents NameError when user writes "global foo" before "foo = ..."
                        for name in user_global_names:
                            if name not in self.namespace:
                                self.namespace[name] = None

                        # Filter to only existing namespace vars (like Jupyter does)
                        # Include both: assigned vars that exist + user's explicit globals
                        existing_vars = {
                            name for name in (assigned_names | user_global_names) if name in self.namespace
                        }
                    except Exception:
                        existing_vars = set()

                    # Build global declaration if needed
                    global_decl = ""
                    has_global_decl = False
                    if existing_vars:
                        vars_str = ", ".join(sorted(existing_vars))
                        global_decl = f"    global {vars_str}\n"
                        has_global_decl = True

                    indented_code = "\n".join("    " + line if line.strip() else line for line in code.split("\n"))
                    wrapped_code = f"""async def __code_exec__():
{global_decl}{indented_code}
    # Return locals so we can update the namespace
    return locals()

__code_exec_coro__ = __code_exec__()
"""
                    # Store whether we added a global declaration (needed for error line mapping)
                    self.namespace["_has_global_decl"] = has_global_decl

                    # Compile and execute wrapper at module level
                    compiled_code = compile(wrapped_code, "<code>", "exec")
                    exec(compiled_code, self.namespace, self.namespace)

                    # Get and await the coroutine, then update namespace with new/modified variables
                    coro = self.namespace.get("__code_exec_coro__")
                    if coro:
                        result_locals = await coro
                        # Update namespace with all variables from the function's locals
                        # This makes variable assignments persist across cells
                        if result_locals:
                            for key, value in result_locals.items():
                                if not key.startswith("_"):
                                    self.namespace[key] = value
                                    # Variable info is tracked in "Available" section, no need for verbose inline output

                        # Clean up temporary variables
                        self.namespace.pop("__code_exec_coro__", None)
                        self.namespace.pop("__code_exec__", None)
                else:
                    # No await - execute directly at module level for natural variable scoping
                    # This means x = x + 10 will work without needing 'global x'

                    # Track variables before execution
                    vars_before = set(self.namespace.keys())

                    compiled_code = compile(code, "<code>", "exec")
                    exec(compiled_code, self.namespace, self.namespace)

                    # Track newly created/modified variables (info shown in "Available" section)
                    vars_after = set(self.namespace.keys())
                    vars_after - vars_before

                # Get output
                output_value = sys.stdout.getvalue()
                if output_value:
                    output = output_value

            finally:
                sys.stdout = old_stdout

            # Wait 2 seconds for page to stabilize after code execution
            await asyncio.sleep(0.5)

            # Note: Browser state is now fetched right before LLM call instead of after each execution
            # This reduces unnecessary state fetches for operations that don't affect the browser

            cell.status = ExecutionStatus.SUCCESS
            cell.output = output
            cell.browser_state = None  # Will be captured in next iteration before LLM call

        except Exception as e:
            # Handle EvaluateError specially - JavaScript execution failed
            if isinstance(e, EvaluateError):
                error = str(e)
                cell.status = ExecutionStatus.ERROR
                cell.error = error
                logger.error(f"Code execution error: {error}")

                await asyncio.sleep(1)

                # Browser state will be fetched before next LLM call
                # Return immediately - do not continue executing code
                return output, error, None

            # Handle NameError specially - check for code block variable confusion
            if isinstance(e, NameError):
                error_msg = str(e)
                cell.status = ExecutionStatus.ERROR
                cell.error = error

                # Browser state will be fetched before next LLM call
                await asyncio.sleep(0.5)
                return output, error, None

            # For syntax errors and common parsing errors, show just the error message
            # without the full traceback to keep output clean
            if isinstance(e, SyntaxError):
                error_msg = e.msg if e.msg else str(e)
                error = f"{type(e).__name__}: {error_msg}"

                # Detect common f-string issues with JSON/JavaScript code
                if "unterminated" in error_msg.lower() and "string" in error_msg.lower() and code:
                    # Check if code contains f-strings with potential JSON/JS content
                    has_fstring = bool(re.search(r'\bf["\']', code))
                    has_json_pattern = bool(
                        re.search(r'json\.dumps|"[^"]*\{[^"]*\}[^"]*"|\'[^\']*\{[^\']*\}[^\']*\'', code)
                    )
                    has_js_pattern = bool(re.search(r"evaluate\(|await evaluate", code))

                    if has_fstring and (has_json_pattern or has_js_pattern):
                        error += (
                            "\n\n💡 TIP: Detected f-string with JSON/JavaScript code containing {}.\n"
                            "   Use separate ```js or ```markdown blocks instead of f-strings to avoid escaping issues.\n"
                            "   If your code block needs ``` inside it, wrap with 4+ backticks: ````markdown code`\n"
                        )

                # Detect and provide helpful hints for common string literal errors
                if "unterminated" in error_msg.lower() and "string" in error_msg.lower():
                    # Detect what type of string literal is unterminated
                    is_triple = "triple-quoted" in error_msg.lower()
                    msg_lower = error_msg.lower()

                    # Detect prefix type from error message
                    if "f-string" in msg_lower and "raw" in msg_lower:
                        prefix = "rf or fr"
                        desc = "raw f-string"
                    elif "f-string" in msg_lower:
                        prefix = "f"
                        desc = "f-string"
                    elif "raw" in msg_lower and "bytes" in msg_lower:
                        prefix = "rb or br"
                        desc = "raw bytes"
                    elif "raw" in msg_lower:
                        prefix = "r"
                        desc = "raw string"
                    elif "bytes" in msg_lower:
                        prefix = "b"
                        desc = "bytes"
                    else:
                        prefix = ""
                        desc = "string"

                    # Build hint based on triple-quoted vs single/double quoted
                    if is_triple:
                        if prefix:
                            hint = f"Hint: Unterminated {prefix}'''...''' or {prefix}\"\"\"...\"\" ({desc}). Check for missing closing quotes or unescaped quotes inside."
                        else:
                            hint = "Hint: Unterminated '''...''' or \"\"\"...\"\" detected. Check for missing closing quotes or unescaped quotes inside."
                        hint += "\n      If you need ``` inside your string, use a ````markdown varname` code block with 4+ backticks instead."
                    else:
                        if prefix:
                            hint = f"Hint: Unterminated {prefix}'...' or {prefix}\"...\" ({desc}). Check for missing closing quote or unescaped quotes inside."
                        else:
                            hint = "Hint: Unterminated '...' or \"...\" detected. Check for missing closing quote or unescaped quotes inside the string."
                    error += f"\n{hint}"

                # Show the problematic line from the code
                if e.text:
                    error += f"\n{e.text}"
                elif e.lineno and code:
                    # If e.text is empty, extract the line from the code
                    lines = code.split("\n")
                    if 0 < e.lineno <= len(lines):
                        error += f"\n{lines[e.lineno - 1]}"

            else:
                # For other errors, try to extract useful information
                error_str = str(e)
                error = f"{type(e).__name__}: {error_str}" if error_str else f"{type(e).__name__} occurred"

                # For RuntimeError or other exceptions, try to extract traceback info
                # to show which line in the user's code actually failed
                if hasattr(e, "__traceback__"):
                    # Walk the traceback to find the frame with '<code>' filename
                    tb = e.__traceback__
                    while tb is not None:
                        frame = tb.tb_frame
                        if frame.f_code.co_filename == "<code>":
                            # Found the frame executing user code
                            # Get the line number from the traceback
                            tb.tb_lineno
                            break
                        tb = tb.tb_next

            cell.status = ExecutionStatus.ERROR
            cell.error = error
            logger.error(f"Code execution error: {error}")

            await asyncio.sleep(1)

            # Browser state will be fetched before next LLM call

        return output, error, None

    async def _get_browser_state(self) -> tuple[str, str | None]:
        """Get the current browser state as text with ultra-minimal DOM structure for code agents.

        Returns:
                Tuple of (browser_state_text, screenshot_base64)
        """
        if not self.browser_session or not self.dom_service:
            return "Browser state not available", None

        try:
            # Get full browser state including screenshot if use_vision is enabled
            include_screenshot = True
            state = await self.browser_session.get_browser_state_summary(include_screenshot=include_screenshot)

            # Format browser state with namespace context
            browser_state_text = await format_browser_state_for_llm(
                state=state, namespace=self.namespace, browser_session=self.browser_session
            )

            screenshot = state.screenshot if include_screenshot else None
            return browser_state_text, screenshot

        except Exception as e:
            logger.error(f"Failed to get browser state: {e}")
            return f"Error getting browser state: {e}", None

    def _format_execution_result(
        self, code: str, output: str | None, error: str | None, current_step: int | None = None
    ) -> str:
        """Format the execution result for the LLM (without browser state)."""
        result = []

        # Add step progress header if step number provided
        if current_step is not None:
            progress_header = f"Step {current_step}/{self.max_steps} executed"
            # Add consecutive failure tracking if there are errors
            if error and self._consecutive_errors > 0:
                progress_header += f" | Consecutive failures: {self._consecutive_errors}/{self.max_failures}"
            result.append(progress_header)

        if error:
            result.append(f"Error: {error}")

        if output:
            # Truncate output if too long
            if len(output) > 10000:
                output = output[:9950] + "\n[Truncated after 10000 characters]"
            result.append(f"Output: {output}")
        if len(result) == 0:
            result.append("Executed")
        return "\n".join(result)

    def _is_task_done(self) -> bool:
        """Check if the task is marked as done in the namespace."""
        # Check if 'done' was called by looking for a special marker in namespace
        return self.namespace.get("_task_done", False)

    async def _capture_screenshot(self, step_number: int) -> str | None:
        """Capture and store screenshot for eval tracking."""
        if not self.browser_session:
            return None

        try:
            # Get browser state summary which includes screenshot
            state = await self.browser_session.get_browser_state_summary(include_screenshot=True)
            if state and state.screenshot:
                # Store screenshot using screenshot service
                screenshot_path = await self.screenshot_service.store_screenshot(state.screenshot, step_number)
                return str(screenshot_path) if screenshot_path else None
        except Exception as e:
            logger.warning(f"Failed to capture screenshot for step {step_number}: {e}")
            return None

    async def _add_step_to_complete_history(
        self,
        model_output_code: str,
        full_llm_response: str,
        output: str | None,
        error: str | None,
        screenshot_path: str | None,
    ) -> None:
        """Add a step to complete_history using type-safe models."""
        # Get current browser URL and title for state
        url: str | None = None
        title: str | None = None
        if self.browser_session:
            try:
                url = await self.browser_session.get_current_page_url()
                # Get title from browser
                cdp_session = await self.browser_session.get_or_create_cdp_session()
                result = await cdp_session.cdp_client.send.Runtime.evaluate(
                    params={"expression": "document.title", "returnByValue": True},
                    session_id=cdp_session.session_id,
                )
                title = result.get("result", {}).get("value")
            except Exception as e:
                logger.debug(f"Failed to get browser URL/title for history: {e}")

        # Check if this is a done result
        is_done = self._is_task_done()

        # Get self-reported success from done() call if task is done
        self_reported_success: bool | None = None
        if is_done:
            task_success = self.namespace.get("_task_success")
            self_reported_success = task_success if isinstance(task_success, bool) else None

        # Create result entry using typed model
        result_entry = CodeAgentResult(
            extracted_content=output if output else None,
            error=error if error else None,
            is_done=is_done,
            success=self_reported_success,
        )

        # Create state entry using typed model
        state_entry = CodeAgentState(url=url, title=title, screenshot_path=screenshot_path)

        # Create metadata entry using typed model
        step_end_time = datetime.datetime.now().timestamp()
        metadata_entry = CodeAgentStepMetadata(
            input_tokens=self._last_llm_usage.prompt_tokens if self._last_llm_usage else None,
            output_tokens=self._last_llm_usage.completion_tokens if self._last_llm_usage else None,
            step_start_time=self._step_start_time,
            step_end_time=step_end_time,
        )

        # Create model output entry using typed model (if there's code to track)
        model_output_entry: CodeAgentModelOutput | None = None
        if model_output_code or full_llm_response:
            model_output_entry = CodeAgentModelOutput(
                model_output=model_output_code if model_output_code else "",
                full_response=full_llm_response if full_llm_response else "",
            )

        # Create history entry using typed model
        history_entry = CodeAgentHistory(
            model_output=model_output_entry,
            result=[result_entry],
            state=state_entry,
            metadata=metadata_entry,
            screenshot_path=screenshot_path,  # Keep for backward compatibility
        )

        self.complete_history.append(history_entry)
        await self._demo_mode_log_step(history_entry)

    async def _demo_mode_log(self, message: str, level: str = "info", metadata: dict[str, Any] | None = None) -> None:
        if not (self._demo_mode_enabled and message and self.browser_session):
            return
        try:
            await self.browser_session.send_demo_mode_log(
                message=message,
                level=level,
                metadata=metadata or {},
            )
        except Exception as exc:
            logger.debug(f"[DemoMode] Failed to send log: {exc}")

    async def _demo_mode_log_step(self, history_entry: CodeAgentHistory) -> None:
        if not self._demo_mode_enabled:
            return
        step_number = len(self.complete_history)
        result = history_entry.result[0] if history_entry.result else None
        if not result:
            return
        level = "error" if result.error else "success" if result.success else "info"
        message_parts = [f"Step {step_number}:"]
        if result.error:
            message_parts.append(f"Error: {result.error}")
        if result.extracted_content:
            message_parts.append(result.extracted_content)
        elif result.success:
            message_parts.append("Marked done.")
        else:
            message_parts.append("Executed.")
        await self._demo_mode_log(
            " ".join(message_parts).strip(),
            level,
            {"step": step_number, "url": history_entry.state.url if history_entry.state else None},
        )

    def _add_sample_output_cell(self, final_result: Any | None) -> None:
        if self._sample_output_added or final_result is None:
            return

        sample_content: str | None = None

        def _extract_sample(data: Any) -> Any | None:
            if isinstance(data, list) and data:
                return data[0]
            if isinstance(data, dict) and data:
                first_key = next(iter(data))
                return {first_key: data[first_key]}
            return data if isinstance(data, (str, int, float, bool)) else None

        data: Any | None = None
        if isinstance(final_result, str):
            try:
                data = json.loads(final_result)
            except Exception:
                sample_content = final_result.strip()
        elif isinstance(final_result, (list, dict)):
            data = final_result

        if data is not None:
            sample = _extract_sample(data)
            if isinstance(sample, (dict, list)):
                try:
                    sample_content = json.dumps(sample, indent=2, ensure_ascii=False)
                except Exception:
                    sample_content = str(sample)
            elif sample is not None:
                sample_content = str(sample)

        if not sample_content:
            return

        sample_cell = self.session.add_cell(source="# Sample output preview")
        sample_cell.cell_type = CellType.MARKDOWN
        sample_cell.status = ExecutionStatus.SUCCESS
        sample_cell.execution_count = None
        escaped = html.escape(sample_content)
        sample_cell.output = f"<pre>{escaped}</pre>"

        self._sample_output_added = True

    def screenshot_paths(self, n_last: int | None = None) -> list[str | None]:
        """
        Get screenshot paths from complete_history for eval system.

        Args:
                n_last: Optional number of last screenshots to return

        Returns:
                List of screenshot file paths (or None for missing screenshots)
        """
        paths = [step.screenshot_path for step in self.complete_history]

        if n_last is not None:
            return paths[-n_last:] if len(paths) > n_last else paths

        return paths

    @property
    def message_manager(self) -> Any:
        """
        Compatibility property for eval system.
        Returns a mock object with last_input_messages attribute.
        """

        class MockMessageManager:
            def __init__(self, llm_messages: list[BaseMessage]) -> None:
                # Convert code-use LLM messages to format expected by eval system
                self.last_input_messages = llm_messages

        return MockMessageManager(self._llm_messages)

    @property
    def history(self) -> CodeAgentHistoryList:
        """
        Compatibility property for eval system.
        Returns a CodeAgentHistoryList object with history attribute containing complete_history.
        This is what the eval system expects when it does: agent_history = agent.history
        """
        return CodeAgentHistoryList(self.complete_history, self.usage_summary)

    async def close(self) -> None:
        """Close the browser session."""
        if self.browser_session:
            # Check if we should close the browser based on keep_alive setting
            if not self.browser_session.browser_profile.keep_alive:
                await self.browser_session.kill()
            else:
                logger.debug("Browser keep_alive is True, not closing browser session")

    async def __aenter__(self) -> "CodeAgent":
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type: type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None:
        """Async context manager exit."""
        await self.close()



import asyncio

# Noesium-specific import needed

llm = ChatOpenAI("gpt-4.1-mini")


async def main():
    """
    Main function demonstrating mixed automation with Browser-Use and Playwright.
    """
    print("🚀 Mixed Automation with Browser-Use and Actor API")

    browser = Browser(keep_alive=True)
    await browser.start()

    page = await browser.get_current_page() or await browser.new_page()

    # Go to apple wikipedia page
    await page.goto("https://www.google.com/travel/flights")

    await asyncio.sleep(1)

    round_trip_button = await page.must_get_element_by_prompt("round trip button", llm)
    await round_trip_button.click()

    one_way_button = await page.must_get_element_by_prompt("one way button", llm)
    await one_way_button.click()

    await asyncio.sleep(1)

    agent = Agent(task="Find the cheapest flight from London to Paris on 2025-10-15", llm=llm, browser_session=browser)
    await agent.run()

    input("Press Enter to continue...")

    await browser.stop()


if __name__ == "__main__":
    asyncio.run(main())



"""MCP (Model Context Protocol) client integration for browser-use.

This module provides integration between external MCP servers and browser-use's action registry.
MCP tools are dynamically discovered and registered as browser-use actions.

Example usage:
    from noesium.subagents.bu import Tools
    from noesium.subagents.bu.mcp.client import MCPClient

    tools = Tools()

    # Connect to an MCP server
    mcp_client = MCPClient(
        server_name="my-server",
        command="npx",
        args=["@mycompany/mcp-server@latest"]
    )

    # Register all MCP tools as browser-use actions
    await mcp_client.register_to_tools(tools)

    # Now use with Agent as normal - MCP tools are available as actions
"""

import asyncio
import logging
import time
from typing import Any

from mcp import ClientSession, StdioServerParameters, types
from mcp.client.stdio import stdio_client
from pydantic import BaseModel, ConfigDict, Field, create_model

from ...agent.views import ActionResult
from ...tools.registry.service import Registry
from ...tools.service import Tools

logger = logging.getLogger(__name__)

MCP_AVAILABLE = True


class MCPClient:
    """Client for connecting to MCP servers and exposing their tools as browser-use actions."""

    def __init__(
        self,
        server_name: str,
        command: str,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
    ):
        """Initialize MCP client.

        Args:
                server_name: Name of the MCP server (for logging and identification)
                command: Command to start the MCP server (e.g., "npx", "python")
                args: Arguments for the command (e.g., ["@playwright/mcp@latest"])
                env: Environment variables for the server process
        """
        self.server_name = server_name
        self.command = command
        self.args = args or []
        self.env = env

        self.session: ClientSession | None = None
        self._stdio_task = None
        self._read_stream = None
        self._write_stream = None
        self._tools: dict[str, types.Tool] = {}
        self._registered_actions: set[str] = set()
        self._connected = False
        self._disconnect_event = asyncio.Event()

    async def connect(self) -> None:
        """Connect to the MCP server and discover available tools."""
        if self._connected:
            logger.debug(f"Already connected to {self.server_name}")
            return

        time.time()
        error_msg = None

        try:
            logger.info(f"🔌 Connecting to MCP server '{self.server_name}': {self.command} {' '.join(self.args)}")

            # Create server parameters
            server_params = StdioServerParameters(command=self.command, args=self.args, env=self.env)

            # Start stdio client in background task
            self._stdio_task = asyncio.create_task(self._run_stdio_client(server_params))

            # Wait for connection to be established
            retries = 0
            max_retries = 100  # 10 second timeout (increased for parallel test execution)
            while not self._connected and retries < max_retries:
                await asyncio.sleep(0.1)
                retries += 1

            if not self._connected:
                error_msg = f"Failed to connect to MCP server '{self.server_name}' after {max_retries * 0.1} seconds"
                raise RuntimeError(error_msg)

            logger.info(f"📦 Discovered {len(self._tools)} tools from '{self.server_name}': {list(self._tools.keys())}")

        except Exception as e:
            error_msg = str(e)
            raise

    async def _run_stdio_client(self, server_params: StdioServerParameters):
        """Run the stdio client connection in a background task."""
        try:
            async with stdio_client(server_params) as (read_stream, write_stream):
                self._read_stream = read_stream
                self._write_stream = write_stream

                # Create and initialize session
                async with ClientSession(read_stream, write_stream) as session:
                    self.session = session

                    # Initialize the connection
                    await session.initialize()

                    # Discover available tools
                    tools_response = await session.list_tools()
                    self._tools = {tool.name: tool for tool in tools_response.tools}

                    # Mark as connected
                    self._connected = True

                    # Keep the connection alive until disconnect is called
                    await self._disconnect_event.wait()

        except Exception as e:
            logger.error(f"MCP server connection error: {e}")
            self._connected = False
            raise
        finally:
            self._connected = False
            self.session = None

    async def disconnect(self) -> None:
        """Disconnect from the MCP server."""
        if not self._connected:
            return

        time.time()

        try:
            logger.info(f"🔌 Disconnecting from MCP server '{self.server_name}'")

            # Signal disconnect
            self._connected = False
            self._disconnect_event.set()

            # Wait for stdio task to finish
            if self._stdio_task:
                try:
                    await asyncio.wait_for(self._stdio_task, timeout=2.0)
                except TimeoutError:
                    logger.warning(f"Timeout waiting for MCP server '{self.server_name}' to disconnect")
                    self._stdio_task.cancel()
                    try:
                        await self._stdio_task
                    except asyncio.CancelledError:
                        pass

            self._tools.clear()
            self._registered_actions.clear()

        except Exception as e:
            str(e)
            logger.error(f"Error disconnecting from MCP server: {e}")

    async def register_to_tools(
        self,
        tools: Tools,
        tool_filter: list[str] | None = None,
        prefix: str | None = None,
    ) -> None:
        """Register MCP tools as actions in the browser-use tools.

        Args:
                tools: Browser-use tools to register actions to
                tool_filter: Optional list of tool names to register (None = all tools)
                prefix: Optional prefix to add to action names (e.g., "playwright_")
        """
        if not self._connected:
            await self.connect()

        registry = tools.registry

        for tool_name, tool in self._tools.items():
            # Skip if not in filter
            if tool_filter and tool_name not in tool_filter:
                continue

            # Apply prefix if specified
            action_name = f"{prefix}{tool_name}" if prefix else tool_name

            # Skip if already registered
            if action_name in self._registered_actions:
                continue

            # Register the tool as an action
            self._register_tool_as_action(registry, action_name, tool)
            self._registered_actions.add(action_name)

        logger.info(
            f"✅ Registered {len(self._registered_actions)} MCP tools from '{self.server_name}' as browser-use actions"
        )

    def _register_tool_as_action(self, registry: Registry, action_name: str, tool: Any) -> None:
        """Register a single MCP tool as a browser-use action.

        Args:
                registry: Browser-use registry to register action to
                action_name: Name for the registered action
                tool: MCP Tool object with schema information
        """
        # Parse tool parameters to create Pydantic model
        param_fields = {}

        if tool.inputSchema:
            # MCP tools use JSON Schema for parameters
            properties = tool.inputSchema.get("properties", {})
            required = set(tool.inputSchema.get("required", []))

            for param_name, param_schema in properties.items():
                # Convert JSON Schema type to Python type
                param_type = self._json_schema_to_python_type(param_schema, f"{action_name}_{param_name}")

                # Determine if field is required and handle defaults
                if param_name in required:
                    default = ...  # Required field
                else:
                    # Optional field - make type optional and handle default
                    param_type = param_type | None
                    if "default" in param_schema:
                        default = param_schema["default"]
                    else:
                        default = None

                # Add field with description if available
                field_kwargs = {}
                if "description" in param_schema:
                    field_kwargs["description"] = param_schema["description"]

                param_fields[param_name] = (param_type, Field(default, **field_kwargs))

        # Create Pydantic model for the tool parameters
        if param_fields:
            # Create a BaseModel class with proper configuration
            class ConfiguredBaseModel(BaseModel):
                model_config = ConfigDict(extra="forbid", validate_by_name=True, validate_by_alias=True)

            param_model = create_model(f"{action_name}_Params", __base__=ConfiguredBaseModel, **param_fields)
        else:
            # No parameters - create empty model
            param_model = None

        # Determine if this is a browser-specific tool
        tool.name.startswith("browser_") or "page" in tool.name.lower()

        # Set up action filters
        domains = None
        # Note: page_filter has been removed since we no longer use Page objects
        # Browser tools filtering would need to be done via domain filters instead

        # Create async wrapper function for the MCP tool
        # Need to define function with explicit parameters to satisfy registry validation
        if param_model:
            # Type 1: Function takes param model as first parameter
            async def mcp_action_wrapper(params: param_model) -> ActionResult:  # type: ignore[no-redef]
                """Wrapper function that calls the MCP tool."""
                if not self.session or not self._connected:
                    return ActionResult(error=f"MCP server '{self.server_name}' not connected", success=False)

                # Convert pydantic model to dict for MCP call
                tool_params = params.model_dump(exclude_none=True)

                logger.debug(f"🔧 Calling MCP tool '{tool.name}' with params: {tool_params}")

                time.time()
                error_msg = None

                try:
                    # Call the MCP tool
                    result = await self.session.call_tool(tool.name, tool_params)

                    # Convert MCP result to ActionResult
                    extracted_content = self._format_mcp_result(result)

                    return ActionResult(
                        extracted_content=extracted_content,
                        long_term_memory=f"Used MCP tool '{tool.name}' from {self.server_name}",
                    )

                except Exception as e:
                    error_msg = f"MCP tool '{tool.name}' failed: {str(e)}"
                    logger.error(error_msg)
                    return ActionResult(error=error_msg, success=False)

        else:
            # No parameters - empty function signature
            async def mcp_action_wrapper() -> ActionResult:  # type: ignore[no-redef]
                """Wrapper function that calls the MCP tool."""
                if not self.session or not self._connected:
                    return ActionResult(error=f"MCP server '{self.server_name}' not connected", success=False)

                logger.debug(f"🔧 Calling MCP tool '{tool.name}' with no params")

                time.time()
                error_msg = None

                try:
                    # Call the MCP tool with empty params
                    result = await self.session.call_tool(tool.name, {})

                    # Convert MCP result to ActionResult
                    extracted_content = self._format_mcp_result(result)

                    return ActionResult(
                        extracted_content=extracted_content,
                        long_term_memory=f"Used MCP tool '{tool.name}' from {self.server_name}",
                    )

                except Exception as e:
                    error_msg = f"MCP tool '{tool.name}' failed: {str(e)}"
                    logger.error(error_msg)
                    return ActionResult(error=error_msg, success=False)

        # Set function metadata for better debugging
        mcp_action_wrapper.__name__ = action_name
        mcp_action_wrapper.__qualname__ = f"mcp.{self.server_name}.{action_name}"

        # Register the action with browser-use
        description = tool.description or f"MCP tool from {self.server_name}: {tool.name}"

        # Use the registry's action decorator
        registry.action(description=description, param_model=param_model, domains=domains)(mcp_action_wrapper)

        logger.debug(f"✅ Registered MCP tool '{tool.name}' as action '{action_name}'")

    def _format_mcp_result(self, result: Any) -> str:
        """Format MCP tool result into a string for ActionResult.

        Args:
                result: Raw result from MCP tool call

        Returns:
                Formatted string representation of the result
        """
        # Handle different MCP result formats
        if hasattr(result, "content"):
            # Structured content response
            if isinstance(result.content, list):
                # Multiple content items
                parts = []
                for item in result.content:
                    if hasattr(item, "text"):
                        parts.append(item.text)
                    elif hasattr(item, "type") and item.type == "text":
                        parts.append(str(item))
                    else:
                        parts.append(str(item))
                return "\n".join(parts)
            else:
                return str(result.content)
        elif isinstance(result, list):
            # List of content items
            parts = []
            for item in result:
                if hasattr(item, "text"):
                    parts.append(item.text)
                else:
                    parts.append(str(item))
            return "\n".join(parts)
        else:
            # Direct result or unknown format
            return str(result)

    def _json_schema_to_python_type(self, schema: dict, model_name: str = "NestedModel") -> Any:
        """Convert JSON Schema type to Python type.

        Args:
                schema: JSON Schema definition
                model_name: Name for nested models

        Returns:
                Python type corresponding to the schema
        """
        json_type = schema.get("type", "string")

        # Basic type mapping
        type_mapping = {
            "string": str,
            "number": float,
            "integer": int,
            "boolean": bool,
            "array": list,
            "null": type(None),
        }

        # Handle enums (they're still strings)
        if "enum" in schema:
            return str

        # Handle objects with nested properties
        if json_type == "object":
            properties = schema.get("properties", {})
            if properties:
                # Create nested pydantic model for objects with properties
                nested_fields = {}
                required_fields = set(schema.get("required", []))

                for prop_name, prop_schema in properties.items():
                    # Recursively process nested properties
                    prop_type = self._json_schema_to_python_type(prop_schema, f"{model_name}_{prop_name}")

                    # Determine if field is required and handle defaults
                    if prop_name in required_fields:
                        default = ...  # Required field
                    else:
                        # Optional field - make type optional and handle default
                        prop_type = prop_type | None
                        if "default" in prop_schema:
                            default = prop_schema["default"]
                        else:
                            default = None

                    # Add field with description if available
                    field_kwargs = {}
                    if "description" in prop_schema:
                        field_kwargs["description"] = prop_schema["description"]

                    nested_fields[prop_name] = (prop_type, Field(default, **field_kwargs))

                # Create a BaseModel class with proper configuration
                class ConfiguredBaseModel(BaseModel):
                    model_config = ConfigDict(extra="forbid", validate_by_name=True, validate_by_alias=True)

                try:
                    # Create and return nested pydantic model
                    return create_model(model_name, __base__=ConfiguredBaseModel, **nested_fields)
                except Exception as e:
                    logger.error(f"Failed to create nested model {model_name}: {e}")
                    logger.debug(f"Fields: {nested_fields}")
                    # Fallback to basic dict if model creation fails
                    return dict
            else:
                # Object without properties - just return dict
                return dict

        # Handle arrays with specific item types
        if json_type == "array":
            if "items" in schema:
                # Get the item type recursively
                item_type = self._json_schema_to_python_type(schema["items"], f"{model_name}_item")
                # Return properly typed list
                return list[item_type]
            else:
                # Array without item type specification
                return list

        # Get base type for non-object types
        base_type = type_mapping.get(json_type, str)

        # Handle nullable/optional types
        if schema.get("nullable", False) or json_type == "null":
            return base_type | None

        return base_type

    async def __aenter__(self):
        """Async context manager entry."""
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.disconnect()



import asyncio
import re
import shutil
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any

from markdown_pdf import MarkdownPdf, Section
from pydantic import BaseModel, Field

INVALID_FILENAME_ERROR_MESSAGE = "Error: Invalid filename format. Must be alphanumeric with supported extension."
DEFAULT_FILE_SYSTEM_PATH = "browseruse_agent_data"


class FileSystemError(Exception):
    """Custom exception for file system operations that should be shown to LLM"""


class BaseFile(BaseModel, ABC):
    """Base class for all file types"""

    name: str
    content: str = ""

    # --- Subclass must define this ---
    @property
    @abstractmethod
    def extension(self) -> str:
        """File extension (e.g. 'txt', 'md')"""

    def write_file_content(self, content: str) -> None:
        """Update internal content (formatted)"""
        self.update_content(content)

    def append_file_content(self, content: str) -> None:
        """Append content to internal content"""
        self.update_content(self.content + content)

    # --- These are shared and implemented here ---

    def update_content(self, content: str) -> None:
        self.content = content

    def sync_to_disk_sync(self, path: Path) -> None:
        file_path = path / self.full_name
        file_path.write_text(self.content)

    async def sync_to_disk(self, path: Path) -> None:
        file_path = path / self.full_name
        with ThreadPoolExecutor() as executor:
            await asyncio.get_event_loop().run_in_executor(executor, lambda: file_path.write_text(self.content))

    async def write(self, content: str, path: Path) -> None:
        self.write_file_content(content)
        await self.sync_to_disk(path)

    async def append(self, content: str, path: Path) -> None:
        self.append_file_content(content)
        await self.sync_to_disk(path)

    def read(self) -> str:
        return self.content

    @property
    def full_name(self) -> str:
        return f"{self.name}.{self.extension}"

    @property
    def get_size(self) -> int:
        return len(self.content)

    @property
    def get_line_count(self) -> int:
        return len(self.content.splitlines())


class MarkdownFile(BaseFile):
    """Markdown file implementation"""

    @property
    def extension(self) -> str:
        return "md"


class TxtFile(BaseFile):
    """Plain text file implementation"""

    @property
    def extension(self) -> str:
        return "txt"


class JsonFile(BaseFile):
    """JSON file implementation"""

    @property
    def extension(self) -> str:
        return "json"


class CsvFile(BaseFile):
    """CSV file implementation"""

    @property
    def extension(self) -> str:
        return "csv"


class PdfFile(BaseFile):
    """PDF file implementation"""

    @property
    def extension(self) -> str:
        return "pdf"

    def sync_to_disk_sync(self, path: Path) -> None:
        file_path = path / self.full_name
        try:
            md_pdf = MarkdownPdf()
            md_pdf.add_section(Section(self.content))
            md_pdf.save(file_path)
        except Exception as e:
            raise FileSystemError(f"Error: Could not write to file '{self.full_name}'. {str(e)}")

    async def sync_to_disk(self, path: Path) -> None:
        with ThreadPoolExecutor() as executor:
            await asyncio.get_event_loop().run_in_executor(executor, lambda: self.sync_to_disk_sync(path))


class FileSystemState(BaseModel):
    """Serializable state of the file system"""

    files: dict[str, dict[str, Any]] = Field(default_factory=dict)  # full filename -> file data
    base_dir: str
    extracted_content_count: int = 0


class FileSystem:
    """Enhanced file system with in-memory storage and multiple file type support"""

    def __init__(self, base_dir: str | Path, create_default_files: bool = True):
        # Handle the Path conversion before calling super().__init__
        self.base_dir = Path(base_dir) if isinstance(base_dir, str) else base_dir
        self.base_dir.mkdir(parents=True, exist_ok=True)

        # Create and use a dedicated subfolder for all operations
        self.data_dir = self.base_dir / DEFAULT_FILE_SYSTEM_PATH
        if self.data_dir.exists():
            # clean the data directory
            shutil.rmtree(self.data_dir)
        self.data_dir.mkdir(exist_ok=True)

        self._file_types: dict[str, type[BaseFile]] = {
            "md": MarkdownFile,
            "txt": TxtFile,
            "json": JsonFile,
            "csv": CsvFile,
            "pdf": PdfFile,
        }

        self.files = {}
        if create_default_files:
            self.default_files = ["todo.md"]
            self._create_default_files()

        self.extracted_content_count = 0

    def get_allowed_extensions(self) -> list[str]:
        """Get allowed extensions"""
        return list(self._file_types.keys())

    def _get_file_type_class(self, extension: str) -> type[BaseFile] | None:
        """Get the appropriate file class for an extension."""
        return self._file_types.get(extension.lower(), None)

    def _create_default_files(self) -> None:
        """Create default results and todo files"""
        for full_filename in self.default_files:
            name_without_ext, extension = self._parse_filename(full_filename)
            file_class = self._get_file_type_class(extension)
            if not file_class:
                raise ValueError(f"Error: Invalid file extension '{extension}' for file '{full_filename}'.")

            file_obj = file_class(name=name_without_ext)
            self.files[full_filename] = file_obj  # Use full filename as key
            file_obj.sync_to_disk_sync(self.data_dir)

    def _is_valid_filename(self, file_name: str) -> bool:
        """Check if filename matches the required pattern: name.extension"""
        # Build extensions pattern from _file_types
        extensions = "|".join(self._file_types.keys())
        pattern = rf"^[a-zA-Z0-9_\-]+\.({extensions})$"
        return bool(re.match(pattern, file_name))

    def _parse_filename(self, filename: str) -> tuple[str, str]:
        """Parse filename into name and extension. Always check _is_valid_filename first."""
        name, extension = filename.rsplit(".", 1)
        return name, extension.lower()

    def get_dir(self) -> Path:
        """Get the file system directory"""
        return self.data_dir

    def get_file(self, full_filename: str) -> BaseFile | None:
        """Get a file object by full filename"""
        if not self._is_valid_filename(full_filename):
            return None

        # Use full filename as key
        return self.files.get(full_filename)

    def list_files(self) -> list[str]:
        """List all files in the system"""
        return [file_obj.full_name for file_obj in self.files.values()]

    def display_file(self, full_filename: str) -> str | None:
        """Display file content using file-specific display method"""
        if not self._is_valid_filename(full_filename):
            return None

        file_obj = self.get_file(full_filename)
        if not file_obj:
            return None

        return file_obj.read()

    async def read_file(self, full_filename: str, external_file: bool = False) -> str:
        """Read file content using file-specific read method and return appropriate message to LLM"""
        if external_file:
            try:
                try:
                    _, extension = self._parse_filename(full_filename)
                except Exception:
                    return f"Error: Invalid filename format {full_filename}. Must be alphanumeric with a supported extension."
                if extension in ["md", "txt", "json", "csv"]:
                    import anyio

                    async with await anyio.open_file(full_filename, "r") as f:
                        content = await f.read()
                        return f"Read from file {full_filename}.\n<content>\n{content}\n</content>"
                elif extension == "pdf":
                    import pypdf

                    reader = pypdf.PdfReader(full_filename)
                    num_pages = len(reader.pages)
                    MAX_PDF_PAGES = 10
                    extra_pages = num_pages - MAX_PDF_PAGES
                    extracted_text = ""
                    for page in reader.pages[:MAX_PDF_PAGES]:
                        extracted_text += page.extract_text()
                    extra_pages_text = f"{extra_pages} more pages..." if extra_pages > 0 else ""
                    return f"Read from file {full_filename}.\n<content>\n{extracted_text}\n{extra_pages_text}</content>"
                else:
                    return f"Error: Cannot read file {full_filename} as {extension} extension is not supported."
            except FileNotFoundError:
                return f"Error: File '{full_filename}' not found."
            except PermissionError:
                return f"Error: Permission denied to read file '{full_filename}'."
            except Exception:
                return f"Error: Could not read file '{full_filename}'."

        if not self._is_valid_filename(full_filename):
            return INVALID_FILENAME_ERROR_MESSAGE

        file_obj = self.get_file(full_filename)
        if not file_obj:
            return f"File '{full_filename}' not found."

        try:
            content = file_obj.read()
            return f"Read from file {full_filename}.\n<content>\n{content}\n</content>"
        except FileSystemError as e:
            return str(e)
        except Exception:
            return f"Error: Could not read file '{full_filename}'."

    async def write_file(self, full_filename: str, content: str) -> str:
        """Write content to file using file-specific write method"""
        if not self._is_valid_filename(full_filename):
            return INVALID_FILENAME_ERROR_MESSAGE

        try:
            name_without_ext, extension = self._parse_filename(full_filename)
            file_class = self._get_file_type_class(extension)
            if not file_class:
                raise ValueError(f"Error: Invalid file extension '{extension}' for file '{full_filename}'.")

            # Create or get existing file using full filename as key
            if full_filename in self.files:
                file_obj = self.files[full_filename]
            else:
                file_obj = file_class(name=name_without_ext)
                self.files[full_filename] = file_obj  # Use full filename as key

            # Use file-specific write method
            await file_obj.write(content, self.data_dir)
            return f"Data written to file {full_filename} successfully."
        except FileSystemError as e:
            return str(e)
        except Exception as e:
            return f"Error: Could not write to file '{full_filename}'. {str(e)}"

    async def append_file(self, full_filename: str, content: str) -> str:
        """Append content to file using file-specific append method"""
        if not self._is_valid_filename(full_filename):
            return INVALID_FILENAME_ERROR_MESSAGE

        file_obj = self.get_file(full_filename)
        if not file_obj:
            return f"File '{full_filename}' not found."

        try:
            await file_obj.append(content, self.data_dir)
            return f"Data appended to file {full_filename} successfully."
        except FileSystemError as e:
            return str(e)
        except Exception as e:
            return f"Error: Could not append to file '{full_filename}'. {str(e)}"

    async def replace_file_str(self, full_filename: str, old_str: str, new_str: str) -> str:
        """Replace old_str with new_str in file_name"""
        if not self._is_valid_filename(full_filename):
            return INVALID_FILENAME_ERROR_MESSAGE

        if not old_str:
            return "Error: Cannot replace empty string. Please provide a non-empty string to replace."

        file_obj = self.get_file(full_filename)
        if not file_obj:
            return f"File '{full_filename}' not found."

        try:
            content = file_obj.read()
            content = content.replace(old_str, new_str)
            await file_obj.write(content, self.data_dir)
            return f'Successfully replaced all occurrences of "{old_str}" with "{new_str}" in file {full_filename}'
        except FileSystemError as e:
            return str(e)
        except Exception as e:
            return f"Error: Could not replace string in file '{full_filename}'. {str(e)}"

    async def save_extracted_content(self, content: str) -> str:
        """Save extracted content to a numbered file"""
        initial_filename = f"extracted_content_{self.extracted_content_count}"
        extracted_filename = f"{initial_filename}.md"
        file_obj = MarkdownFile(name=initial_filename)
        await file_obj.write(content, self.data_dir)
        self.files[extracted_filename] = file_obj
        self.extracted_content_count += 1
        return f"Extracted content saved to file {extracted_filename} successfully."

    def describe(self) -> str:
        """List all files with their content information using file-specific display methods"""
        DISPLAY_CHARS = 400
        description = ""

        for file_obj in self.files.values():
            # Skip todo.md from description
            if file_obj.full_name == "todo.md":
                continue

            content = file_obj.read()

            # Handle empty files
            if not content:
                description += f"<file>\n{file_obj.full_name} - [empty file]\n</file>\n"
                continue

            lines = content.splitlines()
            line_count = len(lines)

            # For small files, display the entire content
            whole_file_description = (
                f"<file>\n{file_obj.full_name} - {line_count} lines\n<content>\n{content}\n</content>\n</file>\n"
            )
            if len(content) < int(1.5 * DISPLAY_CHARS):
                description += whole_file_description
                continue

            # For larger files, display start and end previews
            half_display_chars = DISPLAY_CHARS // 2

            # Get start preview
            start_preview = ""
            start_line_count = 0
            chars_count = 0
            for line in lines:
                if chars_count + len(line) + 1 > half_display_chars:
                    break
                start_preview += line + "\n"
                chars_count += len(line) + 1
                start_line_count += 1

            # Get end preview
            end_preview = ""
            end_line_count = 0
            chars_count = 0
            for line in reversed(lines):
                if chars_count + len(line) + 1 > half_display_chars:
                    break
                end_preview = line + "\n" + end_preview
                chars_count += len(line) + 1
                end_line_count += 1

            # Calculate lines in between
            middle_line_count = line_count - start_line_count - end_line_count
            if middle_line_count <= 0:
                description += whole_file_description
                continue

            start_preview = start_preview.strip("\n").rstrip()
            end_preview = end_preview.strip("\n").rstrip()

            # Format output
            if not (start_preview or end_preview):
                description += f"<file>\n{file_obj.full_name} - {line_count} lines\n<content>\n{middle_line_count} lines...\n</content>\n</file>\n"
            else:
                description += f"<file>\n{file_obj.full_name} - {line_count} lines\n<content>\n{start_preview}\n"
                description += f"... {middle_line_count} more lines ...\n"
                description += f"{end_preview}\n"
                description += "</content>\n</file>\n"

        return description.strip("\n")

    def get_todo_contents(self) -> str:
        """Get todo file contents"""
        todo_file = self.get_file("todo.md")
        return todo_file.read() if todo_file else ""

    def get_state(self) -> FileSystemState:
        """Get serializable state of the file system"""
        files_data = {}
        for full_filename, file_obj in self.files.items():
            files_data[full_filename] = {"type": file_obj.__class__.__name__, "data": file_obj.model_dump()}

        return FileSystemState(
            files=files_data, base_dir=str(self.base_dir), extracted_content_count=self.extracted_content_count
        )

    def nuke(self) -> None:
        """Delete the file system directory"""
        shutil.rmtree(self.data_dir)

    @classmethod
    def from_state(cls, state: FileSystemState) -> "FileSystem":
        """Restore file system from serializable state at the exact same location"""
        # Create file system without default files
        fs = cls(base_dir=Path(state.base_dir), create_default_files=False)
        fs.extracted_content_count = state.extracted_content_count

        # Restore all files
        for full_filename, file_data in state.files.items():
            file_type = file_data["type"]
            file_info = file_data["data"]

            # Create the appropriate file object based on type
            if file_type == "MarkdownFile":
                file_obj = MarkdownFile(**file_info)
            elif file_type == "TxtFile":
                file_obj = TxtFile(**file_info)
            elif file_type == "JsonFile":
                file_obj = JsonFile(**file_info)
            elif file_type == "CsvFile":
                file_obj = CsvFile(**file_info)
            elif file_type == "PdfFile":
                file_obj = PdfFile(**file_info)
            else:
                # Skip unknown file types
                continue

            # Add to files dict and sync to disk
            fs.files[full_filename] = file_obj
            file_obj.sync_to_disk_sync(fs.data_dir)

        return fs



from dataclasses import dataclass, field
from typing import Any

from bubus import BaseEvent
from cdp_use.cdp.target import TargetID
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, field_serializer

from ..dom.views import DOMInteractedElement, SerializedDOMState

# Known placeholder image data for about:blank pages - a 4x4 white PNG
PLACEHOLDER_4PX_SCREENSHOT = (
    "iVBORw0KGgoAAAANSUhEUgAAAAQAAAAECAIAAAAmkwkpAAAAFElEQVR4nGP8//8/AwwwMSAB3BwAlm4DBfIlvvkAAAAASUVORK5CYII="
)


# Pydantic
class TabInfo(BaseModel):
    """Represents information about a browser tab"""

    model_config = ConfigDict(
        extra="forbid",
        validate_by_name=True,
        validate_by_alias=True,
        populate_by_name=True,
    )

    # Original fields
    url: str
    title: str
    target_id: TargetID = Field(serialization_alias="tab_id", validation_alias=AliasChoices("tab_id", "target_id"))
    parent_target_id: TargetID | None = Field(
        default=None,
        serialization_alias="parent_tab_id",
        validation_alias=AliasChoices("parent_tab_id", "parent_target_id"),
    )  # parent page that contains this popup or cross-origin iframe

    @field_serializer("target_id")
    def serialize_target_id(self, target_id: TargetID, _info: Any) -> str:
        return target_id[-4:]

    @field_serializer("parent_target_id")
    def serialize_parent_target_id(self, parent_target_id: TargetID | None, _info: Any) -> str | None:
        return parent_target_id[-4:] if parent_target_id else None


class PageInfo(BaseModel):
    """Comprehensive page size and scroll information"""

    # Current viewport dimensions
    viewport_width: int
    viewport_height: int

    # Total page dimensions
    page_width: int
    page_height: int

    # Current scroll position
    scroll_x: int
    scroll_y: int

    # Calculated scroll information
    pixels_above: int
    pixels_below: int
    pixels_left: int
    pixels_right: int

    # Page statistics are now computed dynamically instead of stored


@dataclass
class NetworkRequest:
    """Information about a pending network request"""

    url: str
    method: str = "GET"
    loading_duration_ms: float = 0.0  # How long this request has been loading (ms since request started, max 10s)
    resource_type: str | None = None  # e.g., 'Document', 'Stylesheet', 'Image', 'Script', 'XHR', 'Fetch'


@dataclass
class PaginationButton:
    """Information about a pagination button detected on the page"""

    button_type: str  # 'next', 'prev', 'first', 'last', 'page_number'
    backend_node_id: int  # Backend node ID for clicking
    text: str  # Button text/label
    selector: str  # XPath or other selector to locate the element
    is_disabled: bool = False  # Whether the button appears disabled


@dataclass
class BrowserStateSummary:
    """The summary of the browser's current state designed for an LLM to process"""

    # provided by SerializedDOMState:
    dom_state: SerializedDOMState

    url: str
    title: str
    tabs: list[TabInfo]
    screenshot: str | None = field(default=None, repr=False)
    page_info: PageInfo | None = None  # Enhanced page information

    # Keep legacy fields for backward compatibility
    pixels_above: int = 0
    pixels_below: int = 0
    browser_errors: list[str] = field(default_factory=list)
    is_pdf_viewer: bool = False  # Whether the current page is a PDF viewer
    recent_events: str | None = None  # Text summary of recent browser events
    pending_network_requests: list[NetworkRequest] = field(default_factory=list)  # Currently loading network requests
    pagination_buttons: list[PaginationButton] = field(default_factory=list)  # Detected pagination buttons
    closed_popup_messages: list[str] = field(default_factory=list)  # Messages from auto-closed JavaScript dialogs


@dataclass
class BrowserStateHistory:
    """The summary of the browser's state at a past point in time to usse in LLM message history"""

    url: str
    title: str
    tabs: list[TabInfo]
    interacted_element: list[DOMInteractedElement | None] | list[None]
    screenshot_path: str | None = None

    def get_screenshot(self) -> str | None:
        """Load screenshot from disk and return as base64 string"""
        if not self.screenshot_path:
            return None

        import base64
        from pathlib import Path

        path_obj = Path(self.screenshot_path)
        if not path_obj.exists():
            return None

        try:
            with open(path_obj, "rb") as f:
                screenshot_data = f.read()
            return base64.b64encode(screenshot_data).decode("utf-8")
        except Exception:
            return None

    def to_dict(self) -> dict[str, Any]:
        data = {}
        data["tabs"] = [tab.model_dump() for tab in self.tabs]
        data["screenshot_path"] = self.screenshot_path
        data["interacted_element"] = [el.to_dict() if el else None for el in self.interacted_element]
        data["url"] = self.url
        data["title"] = self.title
        return data


class BrowserError(Exception):
    """Browser error with structured memory for LLM context management.

    This exception class provides separate memory contexts for browser actions:
    - short_term_memory: Immediate context shown once to the LLM for the next action
    - long_term_memory: Persistent error information stored across steps
    """

    message: str
    short_term_memory: str | None = None
    long_term_memory: str | None = None
    details: dict[str, Any] | None = None
    while_handling_event: BaseEvent[Any] | None = None

    def __init__(
        self,
        message: str,
        short_term_memory: str | None = None,
        long_term_memory: str | None = None,
        details: dict[str, Any] | None = None,
        event: BaseEvent[Any] | None = None,
    ):
        """Initialize a BrowserError with structured memory contexts.

        Args:
                message: Technical error message for logging and debugging
                short_term_memory: Context shown once to LLM (e.g., available actions, options)
                long_term_memory: Persistent error info stored in agent memory
                details: Additional metadata for debugging
                event: The browser event that triggered this error
        """
        self.message = message
        self.short_term_memory = short_term_memory
        self.long_term_memory = long_term_memory
        self.details = details
        self.while_handling_event = event
        super().__init__(message)

    def __str__(self) -> str:
        if self.details:
            return f"{self.message} ({self.details}) during: {self.while_handling_event}"
        elif self.while_handling_event:
            return f"{self.message} (while handling: {self.while_handling_event})"
        else:
            return self.message


class URLNotAllowedError(BrowserError):
    """Error raised when a URL is not allowed"""



"""Capability data models, provider protocol, and classification enums (RFC-0005)."""

from __future__ import annotations

from enum import Enum
from typing import Any, Protocol, runtime_checkable

from pydantic import BaseModel, Field


class CapabilityType(str, Enum):
    TOOL = "tool"
    MCP_TOOL = "mcp_tool"
    SKILL = "skill"
    AGENT = "agent"
    CLI_AGENT = "cli_agent"


class DeterminismClass(str, Enum):
    DETERMINISTIC = "deterministic"
    STOCHASTIC = "stochastic"
    EXTERNAL = "external"


class SideEffectClass(str, Enum):
    PURE = "pure"
    IDEMPOTENT = "idempotent"
    EFFECTFUL = "effectful"


class LatencyClass(str, Enum):
    REALTIME = "realtime"
    FAST = "fast"
    BATCH = "batch"


STATEFUL_TYPES = frozenset({CapabilityType.AGENT, CapabilityType.CLI_AGENT})


class CapabilityDescriptor(BaseModel):
    """Typed contract describing what a capability can do."""

    capability_id: str
    version: str = "1.0.0"
    capability_type: CapabilityType
    description: str = ""
    input_schema: dict[str, Any] = Field(default_factory=dict)
    output_schema: dict[str, Any] = Field(default_factory=dict)
    determinism: DeterminismClass = DeterminismClass.STOCHASTIC
    side_effects: SideEffectClass = SideEffectClass.PURE
    latency: LatencyClass = LatencyClass.FAST
    tags: list[str] = Field(default_factory=list)

    @property
    def stateful(self) -> bool:
        return self.capability_type in STATEFUL_TYPES


@runtime_checkable
class CapabilityProvider(Protocol):
    """Anything that provides a capability: tool, MCP tool, skill, or agent."""

    @property
    def descriptor(self) -> CapabilityDescriptor: ...

    async def invoke(self, **kwargs: Any) -> Any: ...

    async def health(self) -> bool: ...


class CapabilityQuery(BaseModel):
    """Structured query for capability discovery."""

    capability_id: str | None = None
    version: str | None = None
    capability_type: CapabilityType | None = None
    tag: str | None = None
    determinism: DeterminismClass | None = None
    healthy_only: bool = True



"""
TacitusAgent Module

This module provides advanced research capabilities using LangGraph and LLM integration.
"""

from .agent import TacitusAgent
from .state import ResearchState

__all__ = [
    "TacitusAgent",
    "ResearchState",
]



"""Python-based highlighting system for drawing bounding boxes on screenshots.

This module replaces JavaScript-based highlighting with fast Python image processing
to draw bounding boxes around interactive elements directly on screenshots.
"""

import asyncio
import base64
import io
import logging
import os

from PIL import Image, ImageDraw, ImageFont

from ..dom.views import DOMSelectorMap
from ..observability import observe_debug
from ..utils import time_execution_async

logger = logging.getLogger(__name__)

# Font cache to prevent repeated font loading and reduce memory usage
_FONT_CACHE: dict[tuple[str, int], ImageFont.FreeTypeFont | None] = {}

# Cross-platform font paths
_FONT_PATHS = [
    "/usr/share/fonts/truetype/dejavu/DejaVuSans-Bold.ttf",  # Linux (Debian/Ubuntu)
    "/usr/share/fonts/TTF/DejaVuSans-Bold.ttf",  # Linux (Arch/Fedora)
    "/System/Library/Fonts/Arial.ttf",  # macOS
    "C:\\Windows\\Fonts\\arial.ttf",  # Windows
    "arial.ttf",  # Windows (system path)
    "Arial Bold.ttf",  # macOS alternative
    "/usr/share/fonts/truetype/liberation/LiberationSans-Bold.ttf",  # Linux alternative
]


def get_cross_platform_font(font_size: int) -> ImageFont.FreeTypeFont | None:
    """Get a cross-platform compatible font with caching to prevent memory leaks.

    Args:
        font_size: Size of the font to load

    Returns:
        ImageFont object or None if no system fonts are available
    """
    # Use cache key based on font size
    cache_key = ("system_font", font_size)

    # Return cached font if available
    if cache_key in _FONT_CACHE:
        return _FONT_CACHE[cache_key]

    # Try to load a system font
    font = None
    for font_path in _FONT_PATHS:
        try:
            font = ImageFont.truetype(font_path, font_size)
            break
        except OSError:
            continue

    # Cache the result (even if None) to avoid repeated attempts
    _FONT_CACHE[cache_key] = font
    return font


def cleanup_font_cache() -> None:
    """Clean up the font cache to prevent memory leaks in long-running applications."""
    _FONT_CACHE.clear()


# Color scheme for different element types
ELEMENT_COLORS = {
    "button": "#FF6B6B",  # Red for buttons
    "input": "#4ECDC4",  # Teal for inputs
    "select": "#45B7D1",  # Blue for dropdowns
    "a": "#96CEB4",  # Green for links
    "textarea": "#FF8C42",  # Orange for text areas (was yellow, now more visible)
    "default": "#DDA0DD",  # Light purple for other interactive elements
}

# Element type mappings
ELEMENT_TYPE_MAP = {
    "button": "button",
    "input": "input",
    "select": "select",
    "a": "a",
    "textarea": "textarea",
}


def get_element_color(tag_name: str, element_type: str | None = None) -> str:
    """Get color for element based on tag name and type."""
    # Check input type first
    if tag_name == "input" and element_type:
        if element_type in ["button", "submit"]:
            return ELEMENT_COLORS["button"]

    # Use tag-based color
    return ELEMENT_COLORS.get(tag_name.lower(), ELEMENT_COLORS["default"])


def should_show_index_overlay(element_index: int | None) -> bool:
    """Determine if index overlay should be shown."""
    return element_index is not None


def draw_enhanced_bounding_box_with_text(
    draw,  # ImageDraw.Draw - avoiding type annotation due to PIL typing issues
    bbox: tuple[int, int, int, int],
    color: str,
    text: str | None = None,
    font: ImageFont.FreeTypeFont | None = None,
    element_type: str = "div",
    image_size: tuple[int, int] = (2000, 1500),
    device_pixel_ratio: float = 1.0,
) -> None:
    """Draw an enhanced bounding box with much bigger index containers and dashed borders."""
    x1, y1, x2, y2 = bbox

    # Draw dashed bounding box with pattern: 1 line, 2 spaces, 1 line, 2 spaces...
    dash_length = 4
    gap_length = 8
    line_width = 2

    # Helper function to draw dashed line
    def draw_dashed_line(start_x, start_y, end_x, end_y):
        if start_x == end_x:  # Vertical line
            y = start_y
            while y < end_y:
                dash_end = min(y + dash_length, end_y)
                draw.line([(start_x, y), (start_x, dash_end)], fill=color, width=line_width)
                y += dash_length + gap_length
        else:  # Horizontal line
            x = start_x
            while x < end_x:
                dash_end = min(x + dash_length, end_x)
                draw.line([(x, start_y), (dash_end, start_y)], fill=color, width=line_width)
                x += dash_length + gap_length

    # Draw dashed rectangle
    draw_dashed_line(x1, y1, x2, y1)  # Top
    draw_dashed_line(x2, y1, x2, y2)  # Right
    draw_dashed_line(x2, y2, x1, y2)  # Bottom
    draw_dashed_line(x1, y2, x1, y1)  # Left

    # Draw much bigger index overlay if we have index text
    if text:
        try:
            # Scale font size for appropriate sizing across different resolutions
            img_width, img_height = image_size

            css_width = img_width  # / device_pixel_ratio
            # Much smaller scaling - 1% of CSS viewport width, max 16px to prevent huge highlights
            base_font_size = max(10, min(20, int(css_width * 0.01)))
            # Use shared font loading function with caching
            big_font = get_cross_platform_font(base_font_size)
            if big_font is None:
                big_font = font  # Fallback to original font if no system fonts found

            # Get text size with bigger font
            if big_font:
                bbox_text = draw.textbbox((0, 0), text, font=big_font)
                text_width = bbox_text[2] - bbox_text[0]
                text_height = bbox_text[3] - bbox_text[1]
            else:
                # Fallback for default font
                bbox_text = draw.textbbox((0, 0), text)
                text_width = bbox_text[2] - bbox_text[0]
                text_height = bbox_text[3] - bbox_text[1]

            # Scale padding appropriately for different resolutions
            padding = max(4, min(10, int(css_width * 0.005)))  # 0.3% of CSS width, max 4px
            element_width = x2 - x1
            element_height = y2 - y1

            # Container dimensions
            container_width = text_width + padding * 2
            container_height = text_height + padding * 2

            # Position in top center - for small elements, place further up to avoid blocking content
            # Center horizontally within the element
            bg_x1 = x1 + (element_width - container_width) // 2

            # Simple rule: if element is small, place index further up to avoid blocking icons
            if element_width < 60 or element_height < 30:
                # Small element: place well above to avoid blocking content
                bg_y1 = max(0, y1 - container_height - 5)
            else:
                # Regular element: place inside with small offset
                bg_y1 = y1 + 2

            bg_x2 = bg_x1 + container_width
            bg_y2 = bg_y1 + container_height

            # Center the number within the index box with proper baseline handling
            text_x = bg_x1 + (container_width - text_width) // 2
            # Add extra vertical space to prevent clipping
            text_y = bg_y1 + (container_height - text_height) // 2 - bbox_text[1]  # Subtract top offset

            # Ensure container stays within image bounds
            img_width, img_height = image_size
            if bg_x1 < 0:
                offset = -bg_x1
                bg_x1 += offset
                bg_x2 += offset
                text_x += offset
            if bg_y1 < 0:
                offset = -bg_y1
                bg_y1 += offset
                bg_y2 += offset
                text_y += offset
            if bg_x2 > img_width:
                offset = bg_x2 - img_width
                bg_x1 -= offset
                bg_x2 -= offset
                text_x -= offset
            if bg_y2 > img_height:
                offset = bg_y2 - img_height
                bg_y1 -= offset
                bg_y2 -= offset
                text_y -= offset

            # Draw bigger background rectangle with thicker border
            draw.rectangle([bg_x1, bg_y1, bg_x2, bg_y2], fill=color, outline="white", width=2)

            # Draw white text centered in the index box
            draw.text((text_x, text_y), text, fill="white", font=big_font or font)

        except Exception as e:
            logger.debug(f"Failed to draw enhanced text overlay: {e}")


def draw_bounding_box_with_text(
    draw,  # ImageDraw.Draw - avoiding type annotation due to PIL typing issues
    bbox: tuple[int, int, int, int],
    color: str,
    text: str | None = None,
    font: ImageFont.FreeTypeFont | None = None,
) -> None:
    """Draw a bounding box with optional text overlay."""
    x1, y1, x2, y2 = bbox

    # Draw dashed bounding box
    dash_length = 2
    gap_length = 6

    # Top edge
    x = x1
    while x < x2:
        end_x = min(x + dash_length, x2)
        draw.line([(x, y1), (end_x, y1)], fill=color, width=2)
        draw.line([(x, y1 + 1), (end_x, y1 + 1)], fill=color, width=2)
        x += dash_length + gap_length

    # Bottom edge
    x = x1
    while x < x2:
        end_x = min(x + dash_length, x2)
        draw.line([(x, y2), (end_x, y2)], fill=color, width=2)
        draw.line([(x, y2 - 1), (end_x, y2 - 1)], fill=color, width=2)
        x += dash_length + gap_length

    # Left edge
    y = y1
    while y < y2:
        end_y = min(y + dash_length, y2)
        draw.line([(x1, y), (x1, end_y)], fill=color, width=2)
        draw.line([(x1 + 1, y), (x1 + 1, end_y)], fill=color, width=2)
        y += dash_length + gap_length

    # Right edge
    y = y1
    while y < y2:
        end_y = min(y + dash_length, y2)
        draw.line([(x2, y), (x2, end_y)], fill=color, width=2)
        draw.line([(x2 - 1, y), (x2 - 1, end_y)], fill=color, width=2)
        y += dash_length + gap_length

    # Draw index overlay if we have index text
    if text:
        try:
            # Get text size
            if font:
                bbox_text = draw.textbbox((0, 0), text, font=font)
                text_width = bbox_text[2] - bbox_text[0]
                text_height = bbox_text[3] - bbox_text[1]
            else:
                # Fallback for default font
                bbox_text = draw.textbbox((0, 0), text)
                text_width = bbox_text[2] - bbox_text[0]
                text_height = bbox_text[3] - bbox_text[1]

            # Smart positioning based on element size
            padding = 5
            element_width = x2 - x1
            element_height = y2 - y1
            element_area = element_width * element_height
            index_box_area = (text_width + padding * 2) * (text_height + padding * 2)

            # Calculate size ratio to determine positioning strategy
            size_ratio = element_area / max(index_box_area, 1)

            if size_ratio < 4:
                # Very small elements: place outside in bottom-right corner
                text_x = x2 + padding
                text_y = y2 - text_height
                # Ensure it doesn't go off screen
                text_x = min(text_x, 1200 - text_width - padding)
                text_y = max(text_y, 0)
            elif size_ratio < 16:
                # Medium elements: place in bottom-right corner inside
                text_x = x2 - text_width - padding
                text_y = y2 - text_height - padding
            else:
                # Large elements: place in center
                text_x = x1 + (element_width - text_width) // 2
                text_y = y1 + (element_height - text_height) // 2

            # Ensure text stays within bounds
            text_x = max(0, min(text_x, 1200 - text_width))
            text_y = max(0, min(text_y, 800 - text_height))

            # Draw background rectangle for maximum contrast
            bg_x1 = text_x - padding
            bg_y1 = text_y - padding
            bg_x2 = text_x + text_width + padding
            bg_y2 = text_y + text_height + padding

            # Use white background with thick black border for maximum visibility
            draw.rectangle([bg_x1, bg_y1, bg_x2, bg_y2], fill="white", outline="black", width=2)

            # Draw bold dark text on light background for best contrast
            draw.text((text_x, text_y), text, fill="black", font=font)

        except Exception as e:
            logger.debug(f"Failed to draw text overlay: {e}")


def process_element_highlight(
    element_id: int,
    element,
    draw,
    device_pixel_ratio: float,
    font,
    filter_highlight_ids: bool,
    image_size: tuple[int, int],
) -> None:
    """Process a single element for highlighting."""
    try:
        # Use absolute_position coordinates directly
        if not element.absolute_position:
            return

        bounds = element.absolute_position

        # Scale coordinates from CSS pixels to device pixels for screenshot
        # The screenshot is captured at device pixel resolution, but coordinates are in CSS pixels
        x1 = int(bounds.x * device_pixel_ratio)
        y1 = int(bounds.y * device_pixel_ratio)
        x2 = int((bounds.x + bounds.width) * device_pixel_ratio)
        y2 = int((bounds.y + bounds.height) * device_pixel_ratio)

        # Ensure coordinates are within image bounds
        img_width, img_height = image_size
        x1 = max(0, min(x1, img_width))
        y1 = max(0, min(y1, img_height))
        x2 = max(x1, min(x2, img_width))
        y2 = max(y1, min(y2, img_height))

        # Skip if bounding box is too small or invalid
        if x2 - x1 < 2 or y2 - y1 < 2:
            return

        # Get element color based on type
        tag_name = element.tag_name if hasattr(element, "tag_name") else "div"
        element_type = None
        if hasattr(element, "attributes") and element.attributes:
            element_type = element.attributes.get("type")

        color = get_element_color(tag_name, element_type)

        # Get element index for overlay and apply filtering
        element_index = getattr(element, "element_index", None)
        index_text = None

        if element_index is not None:
            if filter_highlight_ids:
                # Use the meaningful text that matches what the LLM sees
                meaningful_text = element.get_meaningful_text_for_llm()
                # Show ID only if meaningful text is less than 5 characters
                if len(meaningful_text) < 3:
                    index_text = str(element_index)
            else:
                # Always show ID when filter is disabled
                index_text = str(element_index)

        # Draw enhanced bounding box with bigger index
        draw_enhanced_bounding_box_with_text(
            draw, (x1, y1, x2, y2), color, index_text, font, tag_name, image_size, device_pixel_ratio
        )

    except Exception as e:
        logger.debug(f"Failed to draw highlight for element {element_id}: {e}")


@observe_debug(ignore_input=True, ignore_output=True, name="create_highlighted_screenshot")
@time_execution_async("create_highlighted_screenshot")
async def create_highlighted_screenshot(
    screenshot_b64: str,
    selector_map: DOMSelectorMap,
    device_pixel_ratio: float = 1.0,
    viewport_offset_x: int = 0,
    viewport_offset_y: int = 0,
    filter_highlight_ids: bool = True,
) -> str:
    """Create a highlighted screenshot with bounding boxes around interactive elements.

    Args:
        screenshot_b64: Base64 encoded screenshot
        selector_map: Map of interactive elements with their positions
        device_pixel_ratio: Device pixel ratio for scaling coordinates
        viewport_offset_x: X offset for viewport positioning
        viewport_offset_y: Y offset for viewport positioning

    Returns:
        Base64 encoded highlighted screenshot
    """
    try:
        # Decode screenshot
        screenshot_data = base64.b64decode(screenshot_b64)
        image = Image.open(io.BytesIO(screenshot_data)).convert("RGBA")

        # Create drawing context
        draw = ImageDraw.Draw(image)

        # Load font using shared function with caching
        font = get_cross_platform_font(12)
        # If no system fonts found, font remains None and will use default font

        # Process elements sequentially to avoid ImageDraw thread safety issues
        # PIL ImageDraw is not thread-safe, so we process elements one by one
        for element_id, element in selector_map.items():
            process_element_highlight(
                element_id, element, draw, device_pixel_ratio, font, filter_highlight_ids, image.size
            )

        # Convert back to base64
        output_buffer = io.BytesIO()
        try:
            image.save(output_buffer, format="PNG")
            output_buffer.seek(0)
            highlighted_b64 = base64.b64encode(output_buffer.getvalue()).decode("utf-8")

            logger.debug(f"Successfully created highlighted screenshot with {len(selector_map)} elements")
            return highlighted_b64
        finally:
            # Explicit cleanup to prevent memory leaks
            output_buffer.close()
            if "image" in locals():
                image.close()

    except Exception as e:
        logger.error(f"Failed to create highlighted screenshot: {e}")
        # Clean up on error as well
        if "image" in locals():
            image.close()
        # Return original screenshot on error
        return screenshot_b64


async def get_viewport_info_from_cdp(cdp_session) -> tuple[float, int, int]:
    """Get viewport information from CDP session.

    Returns:
        Tuple of (device_pixel_ratio, scroll_x, scroll_y)
    """
    try:
        # Get layout metrics which includes viewport info and device pixel ratio
        metrics = await cdp_session.cdp_client.send.Page.getLayoutMetrics(session_id=cdp_session.session_id)

        # Extract viewport information
        visual_viewport = metrics.get("visualViewport", {})
        css_visual_viewport = metrics.get("cssVisualViewport", {})
        css_layout_viewport = metrics.get("cssLayoutViewport", {})

        # Calculate device pixel ratio
        css_width = css_visual_viewport.get("clientWidth", css_layout_viewport.get("clientWidth", 1280.0))
        device_width = visual_viewport.get("clientWidth", css_width)
        device_pixel_ratio = device_width / css_width if css_width > 0 else 1.0

        # Get scroll position in CSS pixels
        scroll_x = int(css_visual_viewport.get("pageX", 0))
        scroll_y = int(css_visual_viewport.get("pageY", 0))

        return float(device_pixel_ratio), scroll_x, scroll_y

    except Exception as e:
        logger.debug(f"Failed to get viewport info from CDP: {e}")
        return 1.0, 0, 0


@time_execution_async("create_highlighted_screenshot_async")
async def create_highlighted_screenshot_async(
    screenshot_b64: str, selector_map: DOMSelectorMap, cdp_session=None, filter_highlight_ids: bool = True
) -> str:
    """Async wrapper for creating highlighted screenshots.

    Args:
        screenshot_b64: Base64 encoded screenshot
        selector_map: Map of interactive elements
        cdp_session: CDP session for getting viewport info
        filter_highlight_ids: Whether to filter element IDs based on meaningful text

    Returns:
        Base64 encoded highlighted screenshot
    """
    # Get viewport information if CDP session is available
    device_pixel_ratio = 1.0
    viewport_offset_x = 0
    viewport_offset_y = 0

    if cdp_session:
        try:
            device_pixel_ratio, viewport_offset_x, viewport_offset_y = await get_viewport_info_from_cdp(cdp_session)
        except Exception as e:
            logger.debug(f"Failed to get viewport info from CDP: {e}")

    # Create highlighted screenshot with async processing
    final_screenshot = await create_highlighted_screenshot(
        screenshot_b64, selector_map, device_pixel_ratio, viewport_offset_x, viewport_offset_y, filter_highlight_ids
    )

    filename = os.getenv("BROWSER_USE_SCREENSHOT_FILE")
    if filename:

        def _write_screenshot():
            try:
                with open(filename, "wb") as f:
                    f.write(base64.b64decode(final_screenshot))
                logger.debug("Saved screenshot to " + str(filename))
            except Exception as e:
                logger.warning(f"Failed to save screenshot to {filename}: {e}")

        await asyncio.to_thread(_write_screenshot)
    return final_screenshot


# Export the cleanup function for external use in long-running applications
__all__ = ["create_highlighted_screenshot", "create_highlighted_screenshot_async", "cleanup_font_cache"]



"""
Conversation Manager for AskuraAgent - Handles dynamic conversation analysis and flow control.
"""

from typing import List, Optional

try:
    from langchain_core.messages import HumanMessage

    LANGCHAIN_AVAILABLE = True
except ImportError:
    HumanMessage = None
    LANGCHAIN_AVAILABLE = False

from noesium.core.llm import BaseLLMClient
from noesium.core.utils.logging import get_logger

from .models import AskuraConfig, AskuraState, ConversationContext
from .prompts import get_conversation_analysis_prompts, get_response_generation_prompts

logger = get_logger(__name__)


class ConversationManager:
    """Manages dynamic conversation analysis and flow control."""

    def __init__(self, config: AskuraConfig, llm_client: Optional[BaseLLMClient] = None):
        """Initialize the conversation manager."""
        self.config = config
        self.llm = llm_client

    def analyze_conversation_context(self, state: AskuraState, message_depth: int = 3) -> ConversationContext:
        """Analyze conversation context to understand user preferences and conversation flow."""
        if isinstance(self.config.conversation_purpose, str):
            context = ConversationContext(conversation_purpose=self.config.conversation_purpose)
        else:
            context = ConversationContext(conversation_purpose="\n".join(self.config.conversation_purpose))

        if not state.messages:
            logger.warning("No recent messages found")
            return context

        # Analyze user engagement and style
        user_messages = [msg for msg in state.messages[-message_depth * 2 :] if isinstance(msg, HumanMessage)]
        if not user_messages:
            logger.warning("No user messages found")
            state.missing_info = self._get_missing_slots(state)
            return context

        last_user_text = user_messages[-1].content

        try:
            if not self.llm or not isinstance(last_user_text, str):
                raise ValueError("LLM client or last user text is not valid")

            # Prepare recent messages for analysis - optimize for token efficiency
            recent_messages_text = self._format_recent_messages(user_messages[-message_depth:])

            # Get structured prompts for conversation analysis
            system_prompt, user_prompt = get_conversation_analysis_prompts(
                "conversation_context",
                conversation_purpose=context.conversation_purpose,
                recent_messages=recent_messages_text,
            )

            # Use structured completion with retry for reliable analysis
            context = self.llm.structured_completion(
                messages=[{"role": "system", "content": system_prompt}, {"role": "user", "content": user_prompt}],
                response_model=ConversationContext,
                temperature=0.3,
                max_tokens=500,
            )
        except Exception as e:
            logger.warning(f"Error analyzing conversation context using LLM: {e}")

        # Analyze what information we have and what's missing
        state.missing_info = self._get_missing_slots(state)
        return context

    def generate_response(self, state: AskuraState) -> str:
        """Generate contextual responses to guide conversation flow."""
        # This method generates strategic responses that balance natural conversation
        # with purposeful information collection based on missing information slots

        # Get conversation purpose from context or config
        conversation_purpose = (
            state.conversation_context.conversation_purpose
            if state.conversation_context and state.conversation_context.conversation_purpose
            else (
                self.config.conversation_purpose
                if isinstance(self.config.conversation_purpose, str)
                else "\n".join(self.config.conversation_purpose)
            )
        )

        # Get missing information slots for strategic guidance
        self._get_missing_slots(state)
        missing_info_descriptions = []
        for slot in self.config.information_slots:
            if slot.required and not state.extracted_info.get(slot.name):
                missing_info_descriptions.append(f"- {slot.name}: {slot.description}")

        missing_required_slots = (
            "\n".join(missing_info_descriptions)
            if missing_info_descriptions
            else "All key information has been collected"
        )

        system_prompt, user_prompt = get_response_generation_prompts(
            conversation_purpose=conversation_purpose,
            missing_required_slots=missing_required_slots,
            intent_type=state.next_action_plan.intent_type if state.next_action_plan else "casual conversation",
            next_action_reasoning=(
                state.next_action_plan.reasoning
                if state.next_action_plan
                else "Building rapport and guiding conversation naturally toward the purpose"
            ),
            known_slots=str(state.extracted_info) if state.extracted_info else "Nothing specific collected yet",
        )
        utterance = self.llm.completion(
            messages=[{"role": "system", "content": system_prompt}, {"role": "user", "content": user_prompt}],
            temperature=0.7,
            max_tokens=200,
        )
        if isinstance(utterance, str):
            utterance = utterance.strip()
            # Remove surrounding quotes if present
            if utterance.startswith('"') and utterance.endswith('"'):
                utterance = utterance[1:-1]
            elif utterance.startswith("'") and utterance.endswith("'"):
                utterance = utterance[1:-1]
        return utterance

    def _get_missing_slots(self, state: AskuraState) -> dict[str, str]:
        """Get dictionary of missing information with slot names as keys and descriptions as values."""
        missing = {}
        information_slots = state.extracted_info

        # Sort slots by priority (higher priority first)
        for slot in sorted(self.config.information_slots, key=lambda slot: slot.priority, reverse=True):
            if slot.required and not information_slots.get(slot.name):
                missing[f"ask_{slot.name}"] = f"Need to collect {slot.description}"

        return missing

    def _format_recent_messages(self, messages: List[HumanMessage]) -> str:
        """Format recent messages while preserving important context."""
        if not messages:
            return ""

        # Preserve more context while still being efficient
        formatted = []
        for i, msg in enumerate(messages):
            # Keep full short messages, smart truncation for long ones
            if len(msg.content) <= 300:
                content = msg.content
            else:
                # Keep beginning and end for context
                content = msg.content[:200] + "..." + msg.content[-50:]

            role_prefix = "User" if i == len(messages) - 1 else f"U{i+1}"  # Mark most recent
            formatted.append(f"{role_prefix}: {content}")

        return "\n".join(formatted)  # Use newlines for better readability



"""Global constants for NoeAgent"""

from pathlib import Path

# LLM Model Constants
GEMINI_PRO = "google/gemini-2.5-pro"
GEMINI_FLASH = "google/gemini-2.5-flash"

# Embedding Constants
DEFAULT_EMBEDDING_DIMS = 768

# Configuration Constants
NOE_AGENT_HOME = Path.home() / ".noeagent"
DEFAULT_CONFIG_PATH = NOE_AGENT_HOME / "config.json"
CONFIG_VERSION = "1.0"

# Default Directories
NOE_AGENT_LOGS_DIR = NOE_AGENT_HOME / "logs"
NOE_AGENT_MEMORY_DIR = NOE_AGENT_HOME / "memory"
NOE_AGENT_DATA_DIR = NOE_AGENT_HOME / "data"
NOE_AGENT_SESSIONS_DIR = NOE_AGENT_HOME / "sessions"

# Library temp/cache base (toolkits default here when used standalone)
NOE_TMP_BASE = Path("/tmp/noesium")


def get_toolkit_tmp_dir(toolkit_name: str, subdir: str = "") -> str:
    """Return toolkit temp dir under NOE_TMP_BASE, optionally with a subdir."""
    base = NOE_TMP_BASE / toolkit_name
    if subdir:
        base = base / subdir
    return str(base)



"""
Agent Adapter for Noesium to browser-use

This module provides adapters to integrate browser-use agents with noesium's agent framework.
"""

import asyncio
from typing import Any, Generic, TypeVar

from noesium.core.agent import BaseAgent
from noesium.core.llm import BaseLLMClient
from noesium.core.utils.logging import get_logger
from noesium.subagents.bu import BrowserProfile
from noesium.subagents.bu.adapters.llm_adapter import create_llm_adapter
from noesium.subagents.bu.agent.service import Agent
from noesium.subagents.bu.agent.views import AgentHistoryList

T = TypeVar("T")


class NoesiumAgentAdapter(BaseAgent, Generic[T]):
    """
    Adapter that wraps browser-use Agent for use in noesium's agent framework.

    This adapter provides a noesium-compatible interface to browser-use's
    browser automation capabilities.
    """

    def __init__(
        self,
        llm: BaseLLMClient | None = None,
        browser_profile: BrowserProfile | None = None,
        use_vision: bool = True,
        **kwargs,
    ):
        """
        Initialize the adapter.

        Args:
            llm: LLM client to use. If None, uses default noesium LLM client.
            browser_profile: Browser configuration profile.
            use_vision: Whether to use vision capabilities.
            **kwargs: Additional arguments passed to the underlying Agent.
        """
        super().__init__(llm_provider="openai", model_name=None)

        if llm is None:
            from noesium.core.llm import get_llm_client

            llm = get_llm_client(structured_output=True)

        self.browser_profile = browser_profile or BrowserProfile()
        self.use_vision = use_vision
        self._llm_client = llm
        self._underlying_agent: Agent | None = None

    _logger = None

    @property
    def logger(self):
        """Get the logger instance."""
        if self._logger is None:
            self._logger = get_logger(self.__class__.__name__)
        return self._logger

    @logger.setter
    def logger(self, value):
        """Set the logger instance (for parent class compatibility)."""
        self._logger = value

    async def run(
        self,
        user_message: str,
        context: dict[str, Any] | None = None,
        config: Any = None,
        max_steps: int = 25,
    ) -> AgentHistoryList[T]:
        """
        Run the browser automation agent.

        Args:
            user_message: The task description in natural language.
            context: Optional context dictionary (ignored in browser_use).
            config: Optional config object (ignored in browser_use).
            max_steps: Maximum number of steps to take.

        Returns:
            AgentHistoryList containing the execution history and results.
        """
        if self._underlying_agent is None:
            self._underlying_agent = Agent(
                task=user_message,
                llm=create_llm_adapter(self._llm_client),
                browser_profile=self.browser_profile,
                use_vision=self.use_vision,
            )

        return await self._underlying_agent.run(max_steps=max_steps)

    def run_sync(
        self,
        user_message: str,
        context: dict[str, Any] | None = None,
        config: Any = None,
        max_steps: int = 25,
    ) -> AgentHistoryList[T]:
        """
        Synchronous wrapper around run() for easier usage without asyncio.

        Args:
            user_message: The task description in natural language.
            context: Optional context dictionary (ignored in browser_use).
            config: Optional config object (ignored in browser_use).
            max_steps: Maximum number of steps to take.

        Returns:
            AgentHistoryList containing the execution history and results.
        """
        return asyncio.run(self.run(user_message, context, config, max_steps))


def create_agent_adapter(
    llm: BaseLLMClient | None = None,
    browser_profile: BrowserProfile | None = None,
    use_vision: bool = True,
    **kwargs,
) -> NoesiumAgentAdapter:
    """
    Create a noesium-compatible agent adapter from browser-use Agent.

    Args:
        llm: A noesium BaseLLMClient instance
        browser_profile: Browser configuration profile
        use_vision: Whether to use vision capabilities
        **kwargs: Additional arguments

    Returns:
        NoesiumAgentAdapter instance
    """
    return NoesiumAgentAdapter(llm, browser_profile, use_vision, **kwargs)



"""Domain event base class and standard event types (RFC-1001 Section 6.2)."""

from __future__ import annotations

from abc import abstractmethod
from typing import Any

from pydantic import BaseModel

from .envelope import AgentRef, EventEnvelope, TraceContext


class DomainEvent(BaseModel):
    """Typed domain event that produces an EventEnvelope."""

    @abstractmethod
    def event_type(self) -> str: ...

    def payload(self) -> dict[str, Any]:
        return self.model_dump()

    def to_envelope(
        self,
        producer: AgentRef,
        trace: TraceContext,
        causation_id: str | None = None,
        correlation_id: str | None = None,
    ) -> EventEnvelope:
        return EventEnvelope(
            event_type=self.event_type(),
            producer=producer,
            trace=trace,
            causation_id=causation_id,
            correlation_id=correlation_id,
            payload=self.payload(),
        )


# --- Agent lifecycle ---


class AgentStarted(DomainEvent):
    agent_id: str
    agent_type: str

    def event_type(self) -> str:
        return "agent.started"


class AgentStopped(DomainEvent):
    agent_id: str
    reason: str = ""

    def event_type(self) -> str:
        return "agent.stopped"


# --- Kernel node lifecycle ---


class NodeEntered(DomainEvent):
    node_id: str
    graph_id: str

    def event_type(self) -> str:
        return "kernel.node.entered"


class NodeCompleted(DomainEvent):
    node_id: str
    graph_id: str
    duration_ms: float

    def event_type(self) -> str:
        return "kernel.node.completed"


class CheckpointCreated(DomainEvent):
    checkpoint_id: str
    node_id: str

    def event_type(self) -> str:
        return "kernel.checkpoint.created"


# --- Capability lifecycle ---


class CapabilityRegistered(DomainEvent):
    capability_id: str
    version: str
    agent_id: str
    capability_type: str = "tool"

    def event_type(self) -> str:
        return "capability.registered"


class CapabilityUnregistered(DomainEvent):
    capability_id: str
    version: str
    reason: str = ""

    def event_type(self) -> str:
        return "capability.unregistered"


class CapabilityRequested(DomainEvent):
    """Emitted before resolving a capability (agent invocation path)."""

    caller_agent_id: str
    capability_id: str
    correlation_id: str = ""

    def event_type(self) -> str:
        return "capability.requested"


class CapabilityInvoked(DomainEvent):
    caller_agent_id: str
    target_agent_id: str
    capability_id: str
    capability_type: str = ""
    correlation_id: str = ""

    def event_type(self) -> str:
        return "capability.invoked"


class CapabilityCompleted(DomainEvent):
    capability_id: str
    caller_agent_id: str
    capability_type: str = ""
    correlation_id: str = ""
    result: Any = None
    error: str | None = None
    duration_ms: int = 0

    def event_type(self) -> str:
        return "capability.completed"


class HealthChanged(DomainEvent):
    """Emitted when a stateful provider's health status changes."""

    capability_id: str
    agent_id: str
    healthy: bool

    def event_type(self) -> str:
        return "capability.health_changed"


# --- Memory ---


class MemoryWritten(DomainEvent):
    key: str
    value_type: str
    value: Any = None
    content_type: str = "fact"
    provider_id: str = ""

    def event_type(self) -> str:
        return "memory.written"


# --- Task delegation ---


class TaskRequested(DomainEvent):
    task_id: str
    capability_id: str
    task_payload: dict[str, Any] = {}

    def event_type(self) -> str:
        return "task.requested"


class TaskCompleted(DomainEvent):
    task_id: str
    result: Any = None
    error: str | None = None

    def event_type(self) -> str:
        return "task.completed"


# --- Errors ---


class ErrorOccurred(DomainEvent):
    error_type: str
    message: str
    original_event_id: str | None = None
    stack_trace: str | None = None

    def event_type(self) -> str:
        return "system.error.occurred"
